import numpy as np
from faspy.interestrate.rmp_stats import rmpcdf, rmpppf
import sympy as sy

# %%
def value_at_risk(assets, correlations):
    np_assets = np.array(assets)
    np_corr = np.array(correlations)
    value = np_assets.dot(np_corr)
    return float(np.sqrt(value.dot(np.transpose(np_assets))))


def logreturns(prices, period=1):
    """
    Calculate the log returns of prices
        Parameters:
            prices: numpy.array. An array of prices. Variables are column based
            period: int. number of observation period to calulate the returns
        Returns:
            numpy.array. An array of log returns

    """
    rows, columns = prices.shape
    # slicing arrays
    price_current = prices[: rows - period]
    price_last = prices[period: ]
    returns = np.divide(price_current, price_last)
    log_returns = np.log(returns)
    return log_returns


def volatilities(data):
    return np.std(data, axis=0)


def vol_from_rates(rate, period=1):
    rows, columns = rate.shape
    # slicing arrays
    price_current = rate[: rows - period]
    price_last = rate[period:]
    result = price_current - price_last
    return np.std(result, axis=0)


def returns_from_rates(data, period=1):
    rows, columns = data.shape
    price_current = data[: rows - period]
    price_last = data[period:]
    result = price_current - price_last
    return result


def returns_from_prices(data, period=1):
    rows, columns = data.shape
    price_current = data[: rows - period]
    price_last = data[period:]
    returns = np.divide(price_current, price_last)
    log_returns = np.log(returns)
    return log_returns


def correlation(data):
    return np.corrcoef(data, rowvar=False)


def get_cfweight(point1, point2, point):
    weight = sy.Symbol('weight')

    expr = weight * point1 + (1 - weight) * point2
    solved_rates = sy.solveset(expr - point, weight, domain=sy.S.Reals)
    return solved_rates


def get_cfalloc(vol1, vol2, vol, cor1_2):

    alpha = sy.Symbol('alpha')
    # a = sy.Symbol('a')
    # b = sy.Symbol('b')
    # c = sy.Symbol('c')
    a = vol1**2 + vol2**2 - 2*cor1_2*vol1*vol2
    b = 2 * cor1_2 * vol1 * vol2 - 2 * vol2 ** 2
    c = vol2**2 - vol**2

    expr = a*alpha**2 + b*alpha + c
    solved_rates = sy.solveset(expr, alpha, domain=sy.S.Reals)
    return solved_rates


# mapping of cash flow to vertices is only applicable to forward products
# hence the requirement for discount factors
def map_cf_to_var_vertices(cashflows, disfac, var_vertices, corr, vola):
    # cashflows is an array of dictionary [{"times": float, "pv": float},]
    cf_vertices = list(map(lambda x: 0, var_vertices))

    for cf in cashflows:
        booltime = list(map(lambda x: x > cf["times"], var_vertices))
        try:
            index = booltime.index(True)
        except Exception:
            index = 0

        if index == 0:
            weight = 1
            vol = vola[index]
            cf_vertices[index] = cf_vertices[index] + cf['pv']
        else:
            point1, point2 = disfac[index - 1], disfac[index]
            weight = get_cfweight(point1, point2, cf['df'])
            weight = list(weight)
            weight = float(weight[0])

            vol1, vol2 = vola[index-1], vola[index]
            vol = weight * vol1+ (1-weight)* vol2
            cor1_2 = corr[index-1][index]

            cf_alloc = get_cfalloc(vol1,vol2,vol,cor1_2)
            cf_alloc = list(cf_alloc)

            cf_alloc = list(map(lambda x: float(x),cf_alloc))
            cf_alloc = list(filter(lambda x: x>=0 and x <=1,cf_alloc))
            #>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
            #The following need to be check
            if len(cf_alloc)==0:
                cf_alloc = 1
            else:
                cf_alloc = cf_alloc[0]
            #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            cf_vertices[index-1] += cf_alloc * cf['pv']
            cf_vertices[index] += (1-cf_alloc) * cf['pv']
    return cf_vertices


def var_asset_weightbyprice(assets, confidence, vol, fxrate=1):
    np_assets = np.array(assets) * np.array(vol)
    ppf = rmpppf(confidence)
    return np_assets * ppf * fxrate


def var_asset_weightbyrate(assets, confidence, vol, bpvs, fxrate=1):
    # assume rates used for vol calculation is expressed in terms of percentage
    np_assets = np.array(assets) * np.array(vol) * 10000 * np.array(bpvs)
    ppf = rmpppf(confidence)
    return np_assets * ppf * fxrate


def pvbp01_continuousrate(data):
    # data is an array of array [[time, rate]]
    # time is a fraction of a year
    # rate is in terms of percentage
    data_copy = list(data)
    for datum in data_copy:
        pvbp01 = (np.exp(-(datum[0] * datum[1] / 100 + 0.0001)) -
                  np.exp(-(datum[0] * datum[1] / 100)))
        datum.append(pvbp01)

    return data_copy



import numpy as np
from numpy import datetime64 as dt64, timedelta64 as td64
import re
from datetime import datetime
import numba
from .conventions import *
from collections import deque


_a365 = ('Actual/365', 'Act/365', 'A/365', 'English')
_a365f = ('Actual/365 Fixed', 'Actual/365F', 'Act/365 Fixed', 'A/365 Fixed',
          'A/365F')
_a365l = ('Actual/365L', 'Act/365L', 'A/365L')
_a364 = ('Actual/364', 'Act/364', 'A/364')
_a360 = ('Actual/360', 'Act/360', 'A/360', 'French')
_a30_360 = ('30/360 Bond Basis', '30/360 US', '30E/360', '30E/360 ISDA')
_o30_360bb = ('30/360 Bond Basis', '30A/360')
_o30_360us = ('30/360', '30/360 US')
_a30e_360 = ('30E/360', '30/360 ICMA', '30S/360', 'Eurobond basis (ISDA 2006)',
             'Special German')
_a30e_360i = ('30E/360 ISDA', 'Eurobond basis (ISDA 2000)', 'German')
_A_A_icma = ('Actual/Actual', 'Actual/Actual ICMA', 'Act/Act ICMA', 'ISMA-99',
             'Act/Act ISMA')
_A_A_isda = ('Actual/Actual ISDA', 'Actual/365', 'Act/365')
# ,'Actual/Actual','Act/Act')
_A_A_afb = ('Actual/Actual AFB')


@numba.njit('float64(int32, int32, int32, int32, int32, int32)')
def _dcf_360(d1, m1, y1, d2, m2, y2):
    return (360 * (y2 - y1) + 30 * (m2 - m1) + (d2 - d1))/360


def day_count_factor(convention, prev_date, current_date, bondmat_date=None,
                     next_coupon_date=None, business_day=None, Frequency=1):

    if convention in _a365:
        return _dcf_a365(prev_date, current_date)

    elif convention in _a360:
        return _dcf_a360(prev_date, current_date)

    elif convention in _a365f:

        return _dcf_a365f(dt64(prev_date, 'D'), dt64(current_date, 'D'))

    elif convention in _a364:
        return _dcf_a364(prev_date, current_date)

    elif convention in _a365l:
        return _dcf_a365l(prev_date, current_date, Frequency=Frequency)

    elif (convention in _a30_360 or convention in _o30_360bb or
          convention in _o30_360us or convention in _a30e_360 or
          convention in _a30e_360i):
        year1 = np.datetime64(prev_date, 'Y')
        month1 = np.datetime64(prev_date, 'M')
        d1 = np.int32(prev_date - np.datetime64(month1)) + 1
        m1 = np.int32(month1 - np.datetime64(year1, 'M')) + 1
        y1 = np.int64(year1)

        year2 = np.datetime64(current_date, 'Y')
        month2 = np.datetime64(current_date, 'M')
        d2 = np.int32(current_date - np.datetime64(month2)) + 1
        m2 = np.int32(month2 - np.datetime64(year2, 'M')) + 1
        y2 = np.int64(year2)

        if convention in _o30_360bb:
            d1 = min(d1, 30)
            if d1 >= 30:
                d2 = min(d2, 30)
            return _dcf_360(d1, m1, y1, d2, m2, y2)

        elif convention in _o30_360us:
            str_date1 = str(prev_date)
            str_date2 = str(current_date)
            arr_date1 = str_date1.split('-')
            arr_date2 = str_date2.split('-')

            date1m = np.datetime64(prev_date, 'M')
            days = np.int32(prev_date - np.datetime64(date1m, 'D'))
            addmonth = date1m + 1
            nextmonth = np.datetime64(addmonth, 'D') + days
            daysinamonth1 = np.int32(nextmonth - prev_date)

            date2m = np.datetime64(current_date, 'M')
            days2 = np.int32(current_date - np.datetime64(date2m, 'D'))
            addmonth2 = date2m + 1
            nextmonth2 = np.datetime64(addmonth2, 'D') + days2
            daysinamonth2 = np.int32(nextmonth2 - current_date)

            if (np.int32(arr_date1[1]) == 2 and np.int32(arr_date2[1]) == 2
                and days == daysinamonth1 and days2 == daysinamonth2 and
                business_day == 'EOM'):
                d2 = 30
            if (np.int32(arr_date1[1]) == 2 and days == daysinamonth1 and
                business_day == 'End of Month'):
                d1 = 30
            if d2 == 31 and d1 >= 30:
                d2 = 30
            if d1 == 31:
                d1 = 30

            return _dcf_360(d1, m1, y1, d2, m2, y2)

        elif convention in _a30e_360:
            if d1 == 31:
                d1 = 30
            if d2 == 31:
                d2 = 30
            value = _dcf_360(d1, m1, y1, d2, m2, y2)
            return value

        elif convention in _a30e_360i:
            daysinamonth = daysinthemonth(prev_date)
            monthend = np.datetime64(month1, 'D') + daysinamonth - 1
            if monthend == prev_date:
                d1 = 30

            daysinamonth2 = daysinthemonth(current_date)
            monthend2 = np.datetime64(month2, 'D') + daysinamonth2 - 1
            if bondmat_date != monthend2 and m2 != 2:
                d2 = 30

            return _dcf_360(d1, m1, y1, d2, m2, y2)

    elif convention in _A_A_icma:
        if next_coupon_date is None:
            return None
        else:
            accrued_days = _datediff(prev_date, current_date).astype('int')
            coupon_days = _datediff(prev_date, next_coupon_date).astype('int')
            dcf = accrued_days/(Frequency * coupon_days)

            return dcf

    elif convention in _A_A_isda:
        return _dcf_a365(prev_date, current_date, business_day=business_day,
                         Frequency=Frequency)

    elif convention in _A_A_afb:
        return "In development"


def _dcf_a365_4ayear(start, end, business_day=None, Frequency=1):

    year1 = np.datetime64(start, 'Y')
    year2 = np.datetime64(end, 'Y')
    if year2 == year1:
        daysinayear = daysintheyear(start)
        days = np.int64(end - start)
        return days/daysinayear
    else:
        daysinayear1 = daysintheyear(start)
        end1 = np.datetime64(year1 + 1, 'D')
        days1 = np.int64(end1 - start)

        daysinayear2 = daysintheyear(end)
        start2 = np.datetime64(year2, 'D')
        days2 = np.int64(end - start2)
        return days1/daysinayear1 + days2/daysinayear2


def _dcf_a365(prev_date, current_date, business_day=None, Frequency=1):
    year1 = np.datetime64(prev_date, 'Y')
    year2 = np.datetime64(current_date, 'Y')
    years = np.int64(year2 - year1)

    if years == 0:
        return _dcf_a365_4ayear(prev_date, current_date,
                                business_day=business_day,
                                Frequency=Frequency)
    else:
        start_m = np.datetime64(prev_date, 'M')
        start_day = np.int64(np.datetime64(prev_date, 'D') -
                             np.datetime64(start_m, 'D'))

        years_enddate_m = np.datetime64(prev_date, 'M') + years * 12
        years_enddate = np.datetime64(years_enddate_m, 'D') + start_day

        extra_days = np.int64(current_date - years_enddate)

        if years == 1:
            if extra_days == 0:
                return years
            elif extra_days < 0:
                return _dcf_a365_4ayear(prev_date, current_date,
                                        business_day=business_day,
                                        Frequency=Frequency)
            else:
                yearfrac = _dcf_a365_4ayear(years_enddate, current_date,
                                            business_day=business_day,
                                            Frequency=Frequency)
                return years + yearfrac
        else:
            if extra_days == 0:
                return years
            else:
                yearfrac = _dcf_a365_4ayear(years_enddate, current_date,
                                            business_day=business_day,
                                            Frequency=Frequency)
                return years + yearfrac


def _dcf_a365f(prev_date, current_date):

    datediff = _datediff(prev_date, current_date)
    days = datediff.astype('int')
    return days/365


def _dcf_a364(prev_date, current_date, business_day=None):
    days = np.int32(current_date - prev_date)
    return days/364


# Cant use this => @numba.njit('float64(numba.datetime, numba.datetime)')

def _dcf_a360(prev_date, current_date):
    datediff = _datediff(prev_date, current_date)
    days = datediff.astype('int')
    return days/360


def _dcf_a365l(prev_date, current_date, business_day=None, Frequency=1):

    year1 = np.datetime64(prev_date, 'Y')
    year2 = np.datetime64(current_date, 'Y')
    days = np.int32(current_date - prev_date)
    # check for leap year
    if Frequency == 1:
        start = np.datetime64(year1, 'D')
        end = np.datetime64(year1 + 1, 'D')
        daysinayear = np.int32(end - start)
        if daysinayear == 365:
            return days/365
        else:
            leapdate = np.datetime64(year1, 'M') + 1
            leapdate = np.datetime64(leapdate, 'D') + 28
            if prev_date < leapdate and current_date > leapdate:
                return days/366
            else:
                return days/365
    else:
        start2 = np.datetime64(year2, 'D')
        end2 = np.datetime64(year2 + 1, 'D')
        daysinayear2 = np.int32(end2 - start2)

    return days/daysinayear2


def daysinthemonth(date1):
    start = np.datetime64(date1, 'M')
    end = start + 1
    datediff = _datediff(np.datetime64(start, 'D'), np.datetime64(end, 'D'))
    return datediff.astype(int)


def daysintheyear(date1):
    start = np.datetime64(date1, 'Y')
    end = start + 1
    return np.int32(np.datetime64(end, 'D') - np.datetime64(start, 'D'))


def npdate_to_datetime(npdate):
    strdate = str(npdate)
    dtdate = datetime.strptime(strdate, "%Y-%m-%d").date()
    return dtdate


def movedatebyyear(your_date, no_of_year=1):
    start_m = np.datetime64(your_date, 'M')
    end_m = start_m + 12 * no_of_year
    days = np.datetime64(your_date, 'D') - np.datetime64(start_m, 'D')
    enddate = np.datetime64(end_m, 'D') + days
    return enddate


def movedatebymonth(your_date, no_of_month=1, business_day='No Adjustment',
                    holidays=[]):

    bdc = np.busdaycalendar(weekmask='1111100', holidays=holidays)
    start_m = dt64(your_date, 'M')
    day = dt64(your_date, 'D') - dt64(start_m, 'D')
    end_m = start_m + no_of_month
    next_m = end_m + 1
    days_in_end_m = dt64(next_m, 'D') - dt64(end_m, 'D')

    if days_in_end_m > day:
        enddate = dt64(end_m, 'D') + day
    else:
        enddate = dt64(end_m, 'D') + days_in_end_m - 1

    if business_day == 'No Adjustment' or business_day is None:
        return enddate

    elif business_day == 'Following':
        return np.busday_offset(enddate, 0, roll='forward', busdaycal=bdc)

    elif business_day == 'Preceeding':
        return np.busday_offset(enddate, 0, roll='backward', busdaycal=bdc)

    elif business_day == 'Modified Following':
        new_date = np.busday_offset(enddate, 0, roll='forward', busdaycal=bdc)
        if dt64(new_date, 'M') > end_m:
            return np.busday_offset(enddate, 0, roll='backward', busdaycal=bdc)
        else:
            return new_date


def generate_dates(currentDate, maturity, issueDate=None, frequency=6,
                   business_day='No Adjustment',
                   method='Forward from issue date', holidays=[],
                   cpnrate=None, ytm=None):

    dates = deque()

    if method == 'Backward from maturity date':
        cpnDate = maturity
        dates.append(cpnDate)
        counter = 0
        if issueDate is None:
            while currentDate < cpnDate:
                counter += 1
                cpnDate = movedatebymonth(maturity,
                                          no_of_month=-counter * frequency,
                                          business_day=business_day,
                                          holidays=holidays)
                dates.insert(0, cpnDate)

        else:
            if type(issueDate) != np.datetime64:
                raise Exception("issueDate is not a date!")
            while issueDate < cpnDate:
                counter += 1
                cpnDate = movedatebymonth(maturity,
                                          no_of_month=-int(counter * frequency),
                                          business_day=business_day,
                                          holidays=holidays)
                if issueDate >= cpnDate:
                    dates.insert(0, issueDate)
                else:
                    dates.insert(0, cpnDate)

    elif method == 'Forward from issue date':
        if issueDate is None:
            cpnDate = currentDate
            counter = 0
            while maturity > cpnDate:
                counter += 1
                cpnDate = movedatebymonth(currentDate,
                                          no_of_month=int(counter * frequency),
                                          business_day=business_day,
                                          holidays=holidays)
                dates.append(cpnDate)
        else:
            if type(issueDate) != np.datetime64:
                raise Exception("issueDate is not a date!")

            cpnDate = issueDate
            dates.append(cpnDate)
            counter = 0
            while maturity > cpnDate:
                counter += 1
                cpnDate = movedatebymonth(issueDate,
                                          no_of_month=int(counter * frequency),
                                          business_day=business_day,
                                          holidays=holidays)
                if maturity < cpnDate:
                    dates.append(maturity)
                else:
                    dates.append(cpnDate)

    return list(dates)



def generate_datesv102(currentDate, maturity, issueDate=None, frequency=6,
                   business_day='No Adjustment',
                   method='Forward from issue date', holidays=[],
                   cpnrate=None, ytm=None):

    dates = []

    if method == 'Backward from maturity date':
        cpnDate = maturity
        dates.append({'date': cpnDate})
        counter = 0
        if issueDate is None:
            while currentDate < cpnDate:
                counter += 1
                cpnDate = movedatebymonth(maturity,
                                          no_of_month=-counter * frequency,
                                          business_day=business_day,
                                          holidays=holidays)
                dates.insert(0, {'date': cpnDate})

        else:
            if type(issueDate) != np.datetime64:
                raise Exception("issueDate is not a date!")
            while issueDate < cpnDate:
                counter += 1
                cpnDate = movedatebymonth(maturity,
                                          no_of_month=-int(counter * frequency),
                                          business_day=business_day,
                                          holidays=holidays)
                if issueDate >= cpnDate:
                    dates.insert(0, {'date': issueDate})
                else:
                    dates.insert(0, {'date': cpnDate})

    elif method == 'Forward from issue date':
        if issueDate is None:
            cpnDate = currentDate
            counter = 0
            while maturity > cpnDate:
                counter += 1
                cpnDate = movedatebymonth(currentDate,
                                          no_of_month=int(counter * frequency),
                                          business_day=business_day,
                                          holidays=holidays)
                dates.append({'date': cpnDate})
        else:
            if type(issueDate) != np.datetime64:
                raise Exception("issueDate is not a date!")
            cpnDate = issueDate
            dates.append({'date': cpnDate})
            counter = 0
            while maturity > cpnDate:
                counter += 1
                cpnDate = movedatebymonth(issueDate,
                                          no_of_month=int(counter * frequency),
                                          business_day=business_day,
                                          holidays=holidays)
                if maturity < cpnDate:
                    dates.append({'date': maturity})
                else:
                    dates.append({'date': cpnDate})

    return dates

def generate_dates_v101(currentDate, maturity, issueDate=None, frequency=6,
                   business_day='No Adjustment',
                   method='Forward from issue date', holidays=[],
                   cpnrate=None, ytm=None):

    dates = []

    if method == 'Backward from maturity date':

        cpnDate = maturity
        dates.append(cpnDate)
        counter = 0
        if issueDate is None:
            while currentDate < cpnDate:
                counter += 1
                cpnDate = movedatebymonth(maturity,
                                          no_of_month=-counter * frequency,
                                          business_day=business_day,
                                          holidays=holidays)
                dates.insert(0, cpnDate)

        else:
            if type(issueDate) != np.datetime64:
                raise Exception("issueDate is not a date!")
            while issueDate < cpnDate:
                counter += 1
                cpnDate = movedatebymonth(maturity,
                                          no_of_month=-int(counter * frequency),
                                          business_day=business_day,
                                          holidays=holidays)
                if issueDate >= cpnDate:
                    dates.insert(0, issueDate)
                else:
                    dates.insert(0, cpnDate)

    elif method == 'Forward from issue date':
        if issueDate is None:
            cpnDate = currentDate
            counter = 0
            while maturity > cpnDate:
                counter += 1
                cpnDate = movedatebymonth(currentDate,
                                          no_of_month=int(counter * frequency),
                                          business_day=business_day,
                                          holidays=holidays)
                dates.append(cpnDate)
        else:
            if type(issueDate) != np.datetime64:
                raise Exception("issueDate is not a date!")
            cpnDate = issueDate
            dates.append(cpnDate)
            counter = 0
            while maturity > cpnDate:
                counter += 1
                cpnDate = movedatebymonth(issueDate,
                                          no_of_month=int(counter * frequency),
                                          business_day=business_day,
                                          holidays=holidays)
                if maturity < cpnDate:
                    dates.append(maturity)
                else:
                    dates.append(cpnDate)

    return dates



def tenor_to_maturity(start_date, tenors, convention='Actual/365 Fixed',
                      business_day='No Adjustment', holidays=[]):

    bdc = np.busdaycalendar(weekmask='1111100', holidays=holidays)

    if isinstance(tenors, list):
        result = []
        for tenor in tenors:
            mat_date = None

            if 'w' in tenor or 'W' in tenor:
                tenor_len = [float(number) for number in re.findall(r'-?\d+\.?\d*',
                                                                    tenor)]
                if len(tenor_len) == 0:
                    mat_date = None
                    result.append(mat_date)
                else:
                    number = tenor_len[0]
                    number = np.int32(number)
                    mat_date = np.datetime64(start_date, 'D') + 7 * number
                    mat_date = np.busday_offset(mat_date, 0, roll='forward',
                                                busdaycal=bdc)
                    result.append(mat_date)

            elif 'm' in tenor or 'M' in tenor:
                mat_date = mat_tenor_by_month(start_date, tenor,
                                              convention=convention,
                                              business_day=business_day,
                                              holidays=holidays)
                result.append(mat_date)

            elif 'y' in tenor or 'Y' in tenor:
                mat_date = mat_tenor_by_month(start_date, tenor,
                                              convention=convention,
                                              business_day=business_day,
                                              holidays=holidays,
                                              multiplier=12)
                result.append(mat_date)
            elif tenor == 'O/N':
                mat_date = np.busday_offset(start_date, 1, roll='forward',
                                            busdaycal=bdc)
                result.append(mat_date)

        return result

    else:
        result = tenor_to_maturity(start_date, [tenors], convention=convention,
                                   business_day=business_day,
                                   holidays=holidays)
        return result[0]


def mat_tenor_by_month(start_date, tenor, convention=None, business_day=None,
                       holidays=[], multiplier=1):
    tenor_len = [float(number) for number in re.findall(r'-?\d+\.?\d*', tenor)]
    if len(tenor_len) == 0:
        return None
    else:
        number = tenor_len[0]

        try:
            number = np.int32(number) * multiplier
        except:
            return None


        new_date = movedatebymonth(start_date, no_of_month=number,
                                   business_day=business_day,
                                   holidays=holidays)
        return new_date


@numba.njit
def _datediff(date1, date2):
    return date2 - date1


def adjustdates(months, dates, business_day, holidays=[]):

    bdc = np.busdaycalendar(weekmask='1111100', holidays=holidays)
    datalen = len(dates)
    newdates = []

    for i in range(datalen):
        month = months[i]
        dmonth = dt64(dates[i], 'M')
        if dmonth == month:
            if business_day == 'Following':
                dates[i] = np.busday_offset(dates[i], 0, roll='forward',
                                           busdaycal=bdc)

            elif business_day == 'Preceeding':
                dates[i] = np.busday_offset(dates[i], 0, roll='backward',
                                           busdaycal=bdc)

            elif business_day == 'Modified Following':
                new_date = np.busday_offset(dates[i], 0, roll='forward',
                                            busdaycal=bdc)
                if dt64(new_date, 'M') > month:
                    dates[i] = np.busday_offset(dates[i], 0, roll='backward',
                                            busdaycal=bdc)
        elif dmonth > month:
            #print(month,dmonth)
            nextmonth = month + td64(1, 'M')
            nextmonthdate = dt64(nextmonth, 'D')
            days = _datediff(dt64(month, 'D'), nextmonthdate)
            dates[i] = dt64(month, 'D') + days - 1
            if business_day == 'Following':
                dates[i] = np.busday_offset(dates[i], 0, roll='forward',
                                           busdaycal=bdc)

            elif business_day == 'Preceeding':
                dates[i] = np.busday_offset(dates[i], 0, roll='backward',
                                           busdaycal=bdc)

            elif business_day == 'Modified Following':
                new_date = np.busday_offset(dates[i], 0, roll='forward',
                                            busdaycal=bdc)
                if dt64(new_date, 'M') > month:
                    dates[i] = np.busday_offset(dates[i], 0, roll='backward',
                                            busdaycal=bdc)




import numpy as np
from numpy import datetime64 as dt64
from sympy import Symbol, diff, solve
from faspy.interestrate.rmp_dates import generate_dates as gen_dates
from faspy.interestrate.rmp_dates import frequencies as fre
from faspy.interestrate.rmp_dates import day_count_factor as day_cf, convs
import functools
from faspy.interestrate.rmp_solvers import solver_bond_derivatives as sbd
from numba import njit


def init_bond_cfs(face_value, value_date, maturity, convention,
                  freq, cpns=[], issue_date=None, business_day='No Adjustment',
                  date_method='Backward from maturity date',
                  holidays=[], yld=0):

    dates = gen_dates(value_date, maturity, issueDate=issue_date,
                      frequency=fre[freq], business_day=business_day,
                      method=date_method, holidays=holidays)
    return dates


def bond_cfs_values(face_value, value_date, convention, freq='Semi-Annual',
                    cpn_dates=[], cpns=[], business_day='No Adjustment',
                    holidays=[], yld=0):
    if len(cpn_dates) <= len(cpns):
        return None
    elif len(cpn_dates) > len(cpns) + 1:
        return None

    results = {}
    noofdates = len(cpn_dates)
    cpn_dcfs = []
    yld_dcfs = []
    fwd_dfs = []
    fwd_dfs01 = []
    dfs = []
    dfs01 = []
    cpn_interests = []
    times = []
    accrued_interest = 0
    maturity = cpn_dates[-1]
    for i in range(1, noofdates, 1):
        dcf = None
        dcf_unaccr = None
        time = None
        start_date = cpn_dates[i-1]
        end_date = cpn_dates[i]
        # day count factor for coupon interest
        dcf = day_cf(convention, start_date, end_date,
                     bondmat_date=cpn_dates[-1],
                     next_coupon_date=end_date,
                     business_day=business_day,
                     Frequency=12 / fre[freq])
        time = day_cf('Actual/365', value_date, end_date,
                      bondmat_date=cpn_dates[-1],
                      next_coupon_date=end_date,
                      business_day=business_day,
                      Frequency=12 / fre[freq])

        cpn_interest = cpns[i-1] * 0.01 * dcf * face_value
        # calculate day count factor for discount factor
        if start_date < value_date and end_date > value_date:
            dcf_accr = day_cf(convention, start_date, value_date,
                              bondmat_date=maturity,
                              next_coupon_date=end_date,
                              business_day=business_day,
                              Frequency=12 / fre[freq])
            dcf_unaccr = dcf - dcf_accr
            accrued_interest = dcf_accr * face_value * cpns[i-1] * 0.01

        elif end_date <= value_date:
            dcf_unaccr = 0
        else:  # start_date >= value_date:
            dcf_unaccr = dcf

        fwd_df = 1 / (1 + yld * 0.01 * dcf_unaccr)
        fwd_df01 = 1/(1 + (yld + 0.01) * 0.01 * dcf_unaccr)

        times.append(time)
        fwd_dfs.append(fwd_df)
        fwd_dfs01.append(fwd_df01)
        cpn_dcfs.append(dcf)
        yld_dcfs.append(dcf_unaccr)
        cpn_interests.append(cpn_interest)

    results['cpn_interest'] = cpn_interests
    results['fwd_dfs'] = fwd_dfs
    results['cpn_dcfs'] = cpn_dcfs
    results['yld_dcfs'] = yld_dcfs

    fwd_dfs_len = len(fwd_dfs)
    # calculating df from fwd df
    df = None
    df01 = None
    for i in range(fwd_dfs_len):
        # df= None
        if i == 0:
            df = fwd_dfs[i]
            df01 = fwd_dfs01[i]
        else:
            for k in range(i+1):
                if k == 0:
                    df = fwd_dfs[k]
                    df01 = fwd_dfs01[k]
                else:
                    df = df * fwd_dfs[k]
                    df01 = df01 * fwd_dfs01[k]
        dfs.append(df)
        dfs01.append(df01)

    results['dfs'] = dfs
    results['dfs01'] = dfs01
    # should use reduce
    pv = 0
    pv01 = 0
    dfs_len = len(dfs)
    # calculate the bond value
    for i in range(dfs_len):
        if yld_dcfs[i] != 0:
            pv = pv + cpn_interests[i]*dfs[i]
            pv01 = pv01 + cpn_interests[i]*dfs01[i]

    pv = pv + face_value * dfs[dfs_len - 1]
    pv01 = pv01 + face_value * dfs01[dfs_len - 1]
    results['pv'] = pv
    results['pv01'] = pv01

    # calculate the macaulay duration
    macD = 0
    for i in range(dfs_len):
        if yld_dcfs[i] != 0:
            macD = macD + times[i]*cpn_interests[i]*dfs[i]
    macD = ((macD + times[dfs_len - 1] * face_value *
             dfs[dfs_len - 1]) / pv)

    results['modified duration'] = macD/(1 + yld / 12 / fre[freq])
    results['macaulay duration'] = macD
    results['accrued'] = accrued_interest

    derivatives = solver_bond_duration(face_value, cpns, results['cpn_dcfs'],
                                       results['yld_dcfs'], yld)
    results['duration'] = derivatives['duration']
    results['convexity'] = derivatives['convexity']

    return results


# slow solver but working
def solver_bond_yield(fv, pv, cpns=[], cpn_dcfs=[], yld_dcfs=[]):
    yld = Symbol('yld')
    df = Symbol('df')
    expr = Symbol('expr')
    expr = None
    dcfs_len = len(yld_dcfs)
    for i in range(dcfs_len):
        ci = None
        df = None
        if yld_dcfs[i] != 0:
            ci = cpns[i] * cpn_dcfs[i] * 0.01 * fv

            if df is None:
                df = (1 + yld * yld_dcfs[i] * 0.01)

            for k in range(i):
                if yld_dcfs[k] != 0:
                    df = df * (1 + yld * yld_dcfs[k] * 0.01)

            if expr is None:
                expr = ci/df
            else:
                expr = expr + ci/df

            if i == dcfs_len - 1:
                expr = expr + fv/df - (pv)

    solved_yield = solve(expr, yld)
    new_yields = []
    for i in range(len(solved_yield)):
        try:
            float(solved_yield[i])
        except Exception:
            continue
        new_yields.append(solved_yield[i])

    new_yields = list(filter(lambda x: x > 0, new_yields))
    return new_yields


def solver_bond_duration(fv, cpns=[], cpn_dcfs=[], yld_dcfs=[],
                         my_yield=5):
    yld = Symbol('yld')
    df = Symbol('df')
    expr = Symbol('expr')
    expr = None
    dcfs_len = len(yld_dcfs)
    for i in range(dcfs_len):
        ci = None
        df = None
        if yld_dcfs[i] != 0:
            ci = cpns[i] * cpn_dcfs[i] * 0.01 * fv

            if df is None:
                df = (1 + yld * yld_dcfs[i] * 0.01)

            for k in range(i):
                if yld_dcfs[k] != 0:
                    df = df * (1 + yld * yld_dcfs[k] * 0.01)

            if expr is None:
                expr = ci/df
            else:
                expr = expr + ci/df

            if i == dcfs_len - 1:
                expr = expr + fv/df

    expr_duration = diff(expr, yld)
    duration = expr_duration.evalf(subs={yld: my_yield})
    # print('Duration >>>>',duration)
    expr_convex = diff(expr_duration, yld)
    convexity = expr_convex.evalf(subs={yld: my_yield})
    # print('Convexity >>>>>>',convexity)
    return {'duration': duration, 'convexity': convexity}



def create_bond_dates(bond):
    """
    Create list of coupon dates.

    Parameters
    ----------
    bond : dict
        A dictionary containing the following keys - 'day_count', 'postion',
        'frequency', and 'coupon'.

    Returns
    -------
    dates : list of numpy.datetime64
        list of coupon dates

    """
    bus_day = None
    if bond['business_day'] == 'NULL':
        bus_day = 'No Adjustment'
    else:
        bus_day = bond['business_day']

    start_date = bond.get('issue_date')
    if start_date is not None:
        start_date = dt64(start_date, 'D')
    value_date = bond.get('value_date')

    if value_date is not None:
        value_date = dt64(value_date, 'D')

    if start_date is not None and value_date is not None:
        use_date = start_date
    elif start_date is not None:
        use_date = start_date
    elif value_date is not None:
        use_date = value_date
    else:
        raise Exception('Both value_date and issue_date do not have any value')

    dates = gen_dates(use_date, bond['maturity'], issueDate=start_date,
                      frequency=fre[bond['frequency']],
                      business_day=bus_day, method=bond['date_generation'])

    return dates


def create_bond_full_structures(bond, dates):
    """
    Create full structure from coupon dates.

    Parameters
    ----------
    bond : dict
        A dictionary containing the following keys - 'day_count', 'postion',
        'frequency', and 'coupon'.
    dates : array of numpy.datetime64
        full list of coupon dates.

    Returns
    -------
    dict
        a dictionary with the following keys: 'dates', 'dcfs', 'cfs',
            'times', 'coupons', 'fvs','fvs_flow'.

    """
    day_count = bond.get('day_count')
    if day_count == 'NULL' or day_count is None:
        day_count = 'Actual/Actual'

    face_value = bond.get('position')
    if face_value is None:
        face_value = 100.00
    else:
        try:
            face_value = float(face_value)
        except Exception:
            face_value = 100.00

    full_dates = list(dates)

    dates1 = list(dates)
    dates2 = list(dates)
    dates1.pop(-1)
    dates2.pop(0)
    full_dcfs = list(map(lambda x1, x2: float(day_cf(day_count, x1, x2,
                                               bondmat_date=dates2[-1],
                                               next_coupon_date=x2,
                                               Frequency=12 /
                                               fre[bond['frequency']])),
                         dates1, dates2))
    full_dcfs.insert(0, 0.0)
    full_cpns = list(map(lambda x: float(bond['coupon']), full_dates))
    full_cpns[0] = 0.0
    full_fvs = list(map(lambda x: float(face_value), full_dates))
    full_fvs[0] = 0.0
    full_fvs_flow = list(map(lambda x: 0, full_dates))
    full_fvs_flow[-1] = float(face_value)
    # print(face_value,full_cpns,full_dcfs)
    full_cfs = list(map(lambda cpn, dcf: float(cpn * dcf * 0.01 * face_value),
                        full_cpns, full_dcfs))
    full_times = list(map(lambda x: day_cf('Actual/365', dates[0], x,
                                           bondmat_date=dates2[-1],
                                           next_coupon_date=x), full_dates))
    return {'dates': full_dates, 'dcfs': full_dcfs, 'cfs': full_cfs,
            'times': full_times, 'coupons': full_cpns, 'fvs': full_fvs,
            'fvs_flow': full_fvs_flow}


def create_bond_act_structures(value_date, full_structures,
                               day_count='Actual/Actual', freq='Semi-Annual'):
    """
    Create active structure from full coupon structure.
    An active structure consist of coupon information after value date.

    Parameters
    ----------
    value_date : numpy datetime64
        value date of the bond.
    full_structures : dict
        full_structures must have the following keys:
            'dates', 'dcfs', 'cfs', 'times', 'coupons', 'fvs' and 'fvs_flow'.
    day_count: str
        day count convention and must be in the list of accepted values.
    freq: str
        coupon frequency and must be in the list of accepted values.

    Returns
    -------
    dict
        containing the following information/keys:
            'accrued' -  accrued
            'accrued_interest' - accrued_interest,
            'last_coupon_date' - last_coupon_date,
            'coupons' - an array of coupon rate (%) for each coupon date,
            'coupon_flow' -  array of coupon interest,
            'dates' - coupon dates after value date,
            'fvs' - face value applicable for the coupon date,
            'fvs_flow' -  principal flows after value date if any,
            'times' - time in fraction of a year from value date,
            'dcfs' - day count factor for the coupon period.

    """
    if day_count not in convs:
        # print(day_count, convs)
        raise Exception('day count error in create_bond_active_structures')
    dates = full_structures.get('dates')
    # print('full_structures==>>', full_structures)
    full_dates = full_structures['dates']
    full_coupons = full_structures['coupons']
    full_dcfs = full_structures['dcfs']
    full_cfs = full_structures['cfs']
    # full_times = full_structures['times']
    full_fvs = full_structures['fvs']
    full_fvs_flow = full_structures['fvs_flow']
    # pflow = full_structures['fvs_flow']
    date_bool = list(map(lambda x: value_date < x, dates))
    index = date_bool.index(True)

    active_dates = list(full_dates[index:])
    # print(active_dates)
    active_coupons = list(full_coupons[index:])
    active_dcfs = list(full_dcfs[index:])
    active_cfs = list(full_cfs[index:])
    active_fvs = list(full_fvs[index:])
    active_fvs_flow = list(full_fvs_flow[index:])
    accrued = 0
    last_coupon_date = None
    accrued_interest = 0
    if index >= 0:
        next_date = dt64(full_dates[index])
        last_coupon_date = dt64(full_dates[index - 1])
        accrued_dcf = float(day_cf(day_count, last_coupon_date, value_date,
                             bondmat_date=full_dates[-1],
                             next_coupon_date=next_date,
                             Frequency=12/fre[freq]))
        if value_date < next_date:
            accrued = accrued_dcf
            accrued_interest = float((accrued_dcf * full_coupons[index]
                                * full_fvs[index] * 0.01))
        else:
            last_coupon_date = next_date
            accrued = 0
            accrued_interest = 0
    # recalculating time from value date
    active_times = list(map(lambda x: day_cf('Actual/365', value_date, x,
                                             bondmat_date=active_dates[-1],
                                             next_coupon_date=x),
                            active_dates))

    return {'accrued': accrued, 'accrued_interest': accrued_interest,
            'last_coupon_date': last_coupon_date, 'coupons': active_coupons,
            'coupon_flow': active_cfs,
            'dates': active_dates, 'fvs': active_fvs,
            'fvs_flow': active_fvs_flow, 'times': active_times,
            'dcfs': active_dcfs}

#@njit(parallel=True)
def value_active_structures(value_date, yld, face_value, structure,
                            day_count='Actual/Actual', freq='Semi-Annual'):
    """
    Value an active cash flow structure of a bond.

    Parameters
    ----------
    value_date : numpy.datetime64
        Valuation date
    yld : float
        Yield to maturity of the cash flows
    face_value : float
        Face value of the bond
    structure : dict
        The active cash flow structure. The structure should have keys
         and data structure similar to those return by
         create_bond_act_structures.
    day_count : str, optional
        Day count convention. The default is 'Actual/Actual'.
    freq : TYPE, optional
        Coupon frequecy. The default is 'Semi-Annual'.


    Returns
    -------
    risk_stats : dict
        Keys to the dictionary are 'proceed', 'pvbp01', 'price_per_100',
        'modified duration', 'macaulay duration', 'duration', 'convexity',
        and 'active_structures'.

    """
    try:
        val_date = dt64(value_date)
    except Exception:
        raise Exception('Value date in value_active_structures is not a date')

    active_dates = list(structure.get('dates'))
    active_coupons = list(structure.get('coupons'))
    active_dcfs = list(structure.get('dcfs'))
    active_cfs = list(structure.get('coupon_flow'))
    active_fvs = list(structure.get('fvs'))
    active_fvs_flow = list(structure.get('fvs_flow'))
    last_coupon_date = structure.get('last_coupon_date')
    dfs = []
    dfs01 = []

    totalcf = len(active_cfs)
    accrued_dcf = day_cf(day_count, last_coupon_date, val_date,
                         bondmat_date=active_dates[-1],
                         next_coupon_date=active_dates[0],
                         Frequency=12/fre[freq])
    #dfs, dfs01 = _calc_df(active_cfs, active_dcfs, accrued_dcf, yld, )
    # Calculating the discount factor
    # print(valid_cfs)
    for i in range(totalcf):
        df = 1
        df01 = 1
        for ind in range(i+1):
            if ind == 0:
                df = df * 1 / (1 + yld *
                               (active_dcfs[ind] - accrued_dcf) * 0.01)
                df01 = df01 * 1 / (1 + (yld + 0.01) *
                                   (active_dcfs[ind] - accrued_dcf) * 0.01)
            else:
                df = df * 1 / (1 + yld * active_dcfs[ind] * 0.01)
                df01 = df01 * 1 / (1 + (yld + 0.01) * active_dcfs[ind] * 0.01)
        # print(df)
        dfs.append(df)
        dfs01.append(df01)

    # present value of each coupon flow
    active_cpns_values = list(map(lambda cf, df: cf * df, active_cfs, dfs))
    # present value of each principal flow
    active_pflow_values = list(map(lambda df, flow:  flow * df, dfs,
                                   active_fvs_flow))
    # net present value of all the principal flows
    pflow_value = functools.reduce(lambda a, b: a + b, active_pflow_values)
    # summing the coupon pv to get the value of the bond
    cpn_values = functools.reduce(lambda a, b: a + b, active_cpns_values)
    bond_values = cpn_values + pflow_value

    # Values after one bp shift
    active_cpns_values01 = list(map(lambda cf, df: cf * df, active_cfs, dfs01))

    # present value of each principal flow
    active_pflow_values01 = list(map(lambda df, flow:  flow * df, dfs01,
                                     active_fvs_flow))

    # net present value of all the principal flows
    pflow_value01 = functools.reduce(lambda a, b: a + b, active_pflow_values01)

    # summing the coupon pv to get the value of the bond
    cpn_values01 = functools.reduce(lambda a, b: a + b, active_cpns_values01)
    bond_values01 = cpn_values01 + pflow_value01
    risk_stats = {}
    risk_stats['proceed'] = bond_values
    # present value of 1 bp
    risk_stats['pvbp01'] = bond_values01 - bond_values
    risk_stats['price_per_100'] = ((risk_stats['proceed'] -
                                    structure['accrued_interest']) /
                                   face_value * 100)
    # recalculating time from value date
    active_times = list(map(lambda x: day_cf('Actual/365', val_date, x,
                                             bondmat_date=active_dates[-1],
                                             next_coupon_date=x),
                            active_dates))
    macD = 0
    for i in range(len(active_times)):
        macD = (macD + active_times[i] *
                (active_cfs[i] + active_fvs_flow[i]) *
                dfs[i])

    macD = macD/risk_stats['proceed']

    risk_stats['modified duration'] = macD / (1 + (yld * 0.01)
                                              / (12 / fre[freq]))
    risk_stats['macaulay duration'] = macD

    derivatives = sbd(face_value, active_cfs, active_dcfs,
                      active_fvs_flow, yld)
    risk_stats['duration'] = derivatives['duration']
    risk_stats['convexity'] = derivatives['convexity']
    risk_stats['active_structures'] = {'coupons': active_coupons,
                                       'coupon_flow': active_cfs,
                                       'coupon_values': active_cpns_values,
                                       'dates': active_dates,
                                       'fvs': active_fvs,
                                       'fvs_flow': active_fvs_flow,
                                       'fvs_flow_value': active_pflow_values,
                                       'times': active_times, 'dfs': dfs,
                                       'dcfs': active_dcfs}
    return risk_stats


def value_active_structures_jit(value_date, yld, face_value, structure,
                                day_count='Actual/Actual', freq='Semi-Annual'):
    """
    Value an active cash flow structure of a bond.

    Parameters
    ----------
    value_date : numpy.datetime64
        Valuation date
    yld : float
        Yield to maturity of the cash flows
    face_value : float
        Face value of the bond
    structure : dict
        The active cash flow structure. The structure should have keys
         and data structure similar to those return by
         create_bond_act_structures.
    day_count : str, optional
        Day count convention. The default is 'Actual/Actual'.
    freq : TYPE, optional
        Coupon frequecy. The default is 'Semi-Annual'.


    Returns
    -------
    risk_stats : dict
        Keys to the dictionary are 'proceed', 'pvbp01', 'price_per_100',
        'modified duration', 'macaulay duration', 'duration', 'convexity',
        and 'active_structures'.

    """
    try:
        val_date = dt64(value_date)
    except Exception:
        raise Exception('Value date in value_active_structures is not a date')

    active_dates = np.array(structure.get('dates'))
    active_coupons = np.array(structure.get('coupons'))
    active_dcfs = np.array(structure.get('dcfs'))
    active_cfs = np.array(structure.get('coupon_flow'))
    active_fvs = np.array(structure.get('fvs'))
    active_fvs_flow = np.array(structure.get('fvs_flow'))
    last_coupon_date = structure.get('last_coupon_date')

    # totalcf = len(active_cfs)
    accrued_dcf = day_cf(day_count, last_coupon_date, val_date,
                         bondmat_date=active_dates[-1],
                         next_coupon_date=active_dates[0],
                         Frequency=12/fre[freq])
    dfs, dfs01 = _calc_df(np.array(active_cfs), np.array(active_dcfs),
                          accrued_dcf, yld)

    # print(_calc_risk(active_cfs, dfs, dfs01, active_fvs_flow))
    proceed, pvbp01, active_cpns_values, active_pflow_values = _calc_risk(
        np.array(active_cfs), np.array(dfs), np.array(dfs01),
        np.array(active_fvs_flow))

    risk_stats = {}
    risk_stats['proceed'] = proceed
    # present value of 1 bp
    risk_stats['pvbp01'] = pvbp01
    risk_stats['price_per_100'] = ((risk_stats['proceed'] -
                                    structure['accrued_interest']) /
                                   face_value * 100)
    # recalculating time from value date
    active_times = list(map(lambda x: day_cf('Actual/365', val_date, x,
                                             bondmat_date=active_dates[-1],
                                             next_coupon_date=x),
                            active_dates))
    macD = 0
    for i in range(len(active_times)):
        macD = (macD + active_times[i] *
                (active_cfs[i] + active_fvs_flow[i]) *
                dfs[i])

    macD = macD / risk_stats['proceed']

    risk_stats['modified duration'] = macD / (1 + (yld * 0.01)
                                              / (12 / fre[freq]))
    risk_stats['macaulay duration'] = macD

    derivatives = sbd(face_value, active_cfs, active_dcfs,
                      active_fvs_flow, yld)
    risk_stats['duration'] = derivatives['duration']
    risk_stats['convexity'] = derivatives['convexity']
    risk_stats['active_structures'] = {'coupons': active_coupons,
                                       'coupon_flow': active_cfs,
                                       'coupon_values': active_cpns_values,
                                       'dates': active_dates,
                                       'fvs': active_fvs,
                                       'fvs_flow': active_fvs_flow,
                                       'fvs_flow_value': active_pflow_values,
                                       'times': active_times, 'dfs': dfs,
                                       'dcfs': active_dcfs}
    return risk_stats


@njit
def _calc_df(active_cfs, active_dcfs, accrued_dcf, yld):
    dfs = []
    dfs01 = []
    totalcf = len(active_cfs)
    for i in range(totalcf):
        df = 1
        df01 = 1
        for ind in range(i+1):
            if ind == 0:
                df = df * 1 / (1 + yld *
                               (active_dcfs[ind] - accrued_dcf) * 0.01)
                df01 = df01 * 1 / (1 + (yld + 0.01) *
                                   (active_dcfs[ind] - accrued_dcf) * 0.01)
            else:
                df = df * 1 / (1 + yld * active_dcfs[ind] * 0.01)
                df01 = df01 * 1 / (1 + (yld + 0.01) * active_dcfs[ind] * 0.01)
        # print(df)
        dfs.append(df)
        dfs01.append(df01)

    return (dfs, dfs01)


@njit
def _calc_risk(active_cfs, dfs, dfs01, active_fvs_flow):
    # present value of each coupon flow
    dfs_len = len(dfs)
    active_cpns_values = []
    active_cpns_values01 = []
    active_pflow_values = []
    active_pflow_values01 = []
    for i in range(dfs_len):
        df = dfs[i]
        df01 = dfs01[i]
        cf = active_cfs[i]
        flow = active_fvs_flow[i]
        active_cpns_values.append(float(cf * df))
        active_pflow_values.append(float(flow * df))
        active_cpns_values01.append(float(cf * df01))
        active_pflow_values01.append(float(flow * df01))

    # net present value of all the principal flows
    pflow_value = 0.00
    pflow_len = len(active_pflow_values)
    for i in range(pflow_len):
        pflow_value = pflow_value + float(active_pflow_values[i])

    # summing the coupon pv to get the value of the bond
    cpn_values = 0.00
    cpn_len = len(active_cpns_values)
    for i in range(cpn_len):
        cpn_values = cpn_values + float(active_cpns_values[i])

    bond_values = cpn_values + pflow_value

    # net present value of all the principal flows
    pflow_value01 = 0.00
    pflow_len = len(active_pflow_values01)
    for i in range(pflow_len):
        pflow_value01 = pflow_value01 + float(active_pflow_values01[i])

    # summing the coupon pv to get the value of the bond
    cpn_values01 = 0.00
    cpn_len = len(active_cpns_values01)
    for i in range(cpn_len):
        cpn_values01 = cpn_values01 + float(active_cpns_values01[i])

    bond_values01 = cpn_values01 + pflow_value01
    pvbp01 = float(bond_values01 - bond_values)

    return (bond_values, pvbp01, active_cpns_values, active_pflow_values)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Nov  5 14:08:14 2020

@author: RMS671214
"""

class Level:

    def  __init__(self, description, factor):
        self.__factor = factor
        self.__description = description
        self.__child = {}
        # self.__parent = None
        self.__amount = 0.00

    @property
    def factor(self):
        return self.__factor

    @factor.setter
    def factor(self, factor):
        self.__factor = factor

    @property
    def description(self):
        return self.__description

    @description.setter
    def description(self, description):
        self.__description = description

    @property
    def child(self):
        return self.__child

    @property
    def amount(self):
        return self.__amount

    @amount.setter
    def amount(self, amount):
        self.__amount = amount

    def value(self):
        val = 0
        level_size = len(self.__child)
        if level_size == 0 and self.__factor is not None:
            val += self.__amount * self.__factor

        else:
            for key in self.__child:
                val += self.__child[key].value()
                if isinstance(self.__factor, float):
                    val += self.__amount * self.__factor
        return val


    def get_array(self):

        level_size = len(self.__child)
        if level_size == 0:
            mydict = {}
            mydict["description"] = self.__description
            mydict["amount"] = self.__amount
            mydict["factor"] = self.__factor
            mydict["value"] = self.value()
            return mydict

        else:
            myarr = {"description": self.__description}
            for key in self.__child:
                myarr[key] = self.__child[key].get_array()
            return myarr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov  2 21:02:58 2020

@author: RMS671214
"""

from interestrate.fixincome import floatbond_value, \
    date_structures, floatbond
import numpy as np
from interestrate import rmp_dates as rd
from interestrate import discount_curve as dcurve
from interestrate import rmp_curves as rcurve
import pandas as pd

# %%
# create the float structure

mybond = {}
mybond['issue_date'] = np.datetime64('2018-10-22')
mybond['value_date'] = np.datetime64('2018-10-22')
mybond['maturity'] = np.datetime64('2020-10-22')
mybond['day_count'] = 'Actual/365 Fixed'
mybond['frequency'] = 'Semi-Annual'
mybond['business_day'] = 'No Adjustment'
mybond['date_generation'] = rd.date_gen_method[1]
mybond['face_value'] = 100
mybond["current_coupon"] = 3.65
mybond['spread'] = 1.5
mybond["margin"] = 1
mybond["fixing_basis"] = "Same Day"

# %%

rate = {"value_date": str(mybond['value_date']),
        "st_busday": "Modified Following",
        "st_ratebasis": "Money Market", "st_daycount": "Actual/365",
        "lt_busday": "No Adjustment", "lt_frequency": "Semi-Annual",
        "lt_daycount": "Actual/Actual",
        "rates": {'O/N': 2.30, '1W': 2.35, '1M': 2.45, '3M': 2.55,
                  '6M': 2.65, '12M': 2.75, '1Y': 2.70, '2Y': 2.80, '3Y': 2.90,
                  '5Y': 3.00, '10Y': 3.10, '30Y': 3.25}}


dfs = dcurve.discount_factor_gen(rate, return_type="times")
x_axis = [x["times"] for x in dfs]
y_axis = [x["df"] for x in dfs]
df_func = rcurve.interpolation(x_axis, y_axis, float(1/366), is_function=True)
pd_ori = pd.DataFrame(dfs)

# %%
# Test Date Structure
coupon_dates = date_structures(mybond)

# %%

zcurve = rcurve.discount_factor_from_zspread(mybond['value_date'],
                                             coupon_dates,
                                             mybond['day_count'],
                                             mybond['frequency'],
                                             dfs, mybond['spread'])
pd_z = pd.DataFrame(zcurve)



# %%
fbond = floatbond(mybond, rate)
pd0 = pd.DataFrame(fbond["structure"])
risks = fbond["risks"]
print(fbond["risks"])

#%%

# Using Valuation Curve
val_curve = dict(rate)
val_curve["rates"] = rcurve.shift_curve(val_curve["rates"], bp=mybond["spread"])
fbond2 = floatbond(mybond, rate, val_curve=val_curve)
risks = fbond2["risks"]
print(fbond2["risks"])







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct 30 10:49:32 2020

@author: RMS671214
"""
from .rmp_curves import *
from .fas_ircls import STRates, LTRates, Rates
from .conventions import *
from .rmp_dates import tenor_to_maturity as ttm, generate_dates as gen_dates
from numpy import datetime64 as dt64
import math
from collections import deque

# %%

def discount_factor_gen(rates, return_type="time"):
    """
    Generate the discount factors.

            Parameters:

                rates: a dictionary with the following keys - value_date,
                st_busday, st_ratebasis, st_daycount, lt_busday,
                lt_frequency, lt_daycount and rates. 'rates' key in
                the dictionary is a dictionary of interest rate in
                percentage having the following keys - 1W,  2W, 3W, 1M,
                2M, 3M, 4M, 5M, 6M, 9M, 12M, 1Y, 2Y, 3Y, 4Y, 5Y, 6Y, 7Y,
                10Y, 15Y, 20Y, 30Y.

            Returns:

                a dictionary with the following keys - date, dcf, time,
                days, df, rate
    """
    if isinstance(rates, dict):
        dfs = _discount_factor_generate(rates)
        if return_type == "days":
            df = [{"days": x["days"], "df": x["df"]} for x in dfs]
            df.insert(0, {"days": 0.00, "df": 1.00})
        else:
            df = [{"times": x["time"], "df": x["df"]} for x in dfs]
            df.insert(0, {"times": 0.00, "df": 1.00})
        results = df

    elif isinstance(rates, list):
        results = []
        for rate in rates:
            if isinstance(rate, dict):
                dfs = _discount_factor_generate(rate)
                if return_type == "days":
                    df = [{"days": x["days"], "df": x["df"]} for x in dfs]
                    df.insert(0, {"days": 0.00, "df": 1.00})
                else:
                    df = [{"times": x["time"], "df": x["df"]} for x in dfs]
                    df.insert(0, {"times": 0.00, "df": 1.00})
                results.append(df)

    return results


def _discount_factor_generate(rate):
    strates = STRates()
    strates.rate_basis = rate.get("st_ratebasis")
    strates.business_day = rate.get("st_busday")
    strates.day_count = rate.get("st_daycount")
    strates.rates = rate["rates"]

    ltrates = LTRates()
    ltrates.frequency = rate.get("lt_frequency")
    ltrates.business_day = rate.get("lt_busday")
    ltrates.day_count = rate.get("lt_daycount")
    ltrates.rates = rate["rates"]

    myrates = Rates()
    myrates.strates = strates
    myrates.ltrates = ltrates
    myrates.date_gen_method = 'Forward from issue date'
    myrates.value_date = rate.get("value_date")
    myrates.calcdf()
    return myrates.df.data


def flat_curve(start_date, end_date, rate, rate_basis="Money Market",
               day_count="Actual/365", bus_day="No Adjustment",
               tenors=None, return_type="time"):
    if tenors is None:
        mytenors = [x for x in std_tenors if x != "12M"]
    else:
        mytenors = list(tenors)
    sdate = dt64(start_date, "D")
    edate = dt64(end_date, "D")
    time = day_cf("Actual/365", sdate, edate)
    dcf = day_cf(day_count, sdate, edate)
    if rate_basis == "Money Market":
        df = _mmr2df(float(rate), float(dcf))
    elif rate_basis == "Discount Rate":
        df = _dr2df(float(rate), float(dcf))
    else:
        return None

    time = day_cf("Actual/365", sdate, edate)
    crate = -math.log(df)/time
    dates = [ttm(sdate, tenor, day_count, bus_day) for tenor in mytenors]

    times = [{"times": float(day_cf("Actual/365", sdate, date)) } for date in dates]
    timelen = len(times)
    ret_array=[]
    if return_type == "time":
        dfs = [{"df": math.exp(-crate * time["times"])} for time in times]
        for i in range(timelen):
            time = times[i]
            df = dfs[i]
            time.update(df)
            ret_array.append(time)

    elif return_type == "days":
        days = [{"days": (date-sdate).astype("float") } for date in dates]
        dfs = [{"df": math.exp(-crate * time["times"])} for time in times]
        for i in range(timelen):
            day = days[i]
            df = dfs[i]
            day.update(df)
            ret_array.append(day)
    return ret_array


def discount_factor_from_ytm(value_date, maturity, day_count, frequency,
                             business_day, ytm):
    # dates is a deque
    dates = gen_dates(value_date, maturity, issueDate=value_date,
                      frequency=frequency, business_day=business_day,
                      method=date_gen_method[1])

    start_dates = deque(dates)
    end_dates = deque(dates)
    start_dates.pop()
    end_dates.popleft()
    noofcpns = len(start_dates)
    data = deque()
    for no in range(noofcpns):
        structure = {"start_date": start_dates[no], "end_date": end_dates[no]}
        data.append(structure)

    df_curve =  discount_factor_from_ytm(value_date, list(data), day_count,
                                        frequency, business_day, ytm)
    return df_curve


def discount_factor_from_ytm_using_structures(value_date, date_structure,
                                              day_count,
                                              frequency, business_day, ytm):

    df = 1
    maturity = date_structure[-1]["end_date"]

    data = [{"start_date": x["start_date"], "end_date": x["end_date"]}
            for x in date_structure if value_date < dt64(x["end_date"])]

    for datum in data:
        if dt64(datum["start_date"]) >= value_date:
            datum["dcf"] = day_cf(day_count, datum["start_date"], datum["end_date"],
                                  bondmat_date=maturity,
                                  next_coupon_date=datum["end_date"],
                                  business_day=business_day,
                                  Frequency=12/frequencies[frequency])
            datum["time"] = day_cf("Actual/365", value_date, datum["end_date"],
                                  bondmat_date=maturity,
                                  next_coupon_date=datum["end_date"],
                                  business_day=business_day,
                                  Frequency=12/frequencies[frequency])

        elif datum["start_date"] < value_date:
            datum["dcf"] = day_cf(day_count, value_date, datum["end_date"],
                                  bondmat_date=maturity,
                                  next_coupon_date=datum["end_date"],
                                  business_day=business_day,
                                  Frequency=12/frequencies[frequency])
            datum["time"] = day_cf("Actual/365", value_date, datum["end_date"],
                                  bondmat_date=maturity,
                                  next_coupon_date=datum["end_date"],
                                  business_day=business_day,
                                  Frequency=12/frequencies[frequency])

        datum["period_df"] = 1/(1 + ytm * datum["dcf"] / 100)
        df = datum["df"] = df * datum["period_df"]

    disfac = list(map(lambda datum: {"times": datum["time"], "df": datum["df"]},
                      data))
    disfac.insert(0,{"times":0, "df": 1})
    return disfac


@numba.njit('float64(float64, float64)')
def _df2mmr(df, dcf):
    rate = ((1 / df) - 1) / (0.01 * dcf)
    return rate


@numba.njit('float64(float64, float64)')
def _mmr2df(rate, dcf):
    df = 1 / (1 + rate * 0.01 * dcf)
    return df


@numba.njit('float64(float64, float64)')
def _df2dr(df, dcf):
    rate = (1 - df) / (0.01 * df)
    return rate


@numba.njit('float64(float64, float64)')
def _dr2df(rate, dcf):
    df = 1 - rate * 0.01 * dcf
    return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov  4 07:32:58 2020

@author: RMS671214
"""
from faspy.interestrate import discount_curve as dcurve
from numpy import datetime64 as dt64
vdate = dt64("2020-10-30", "D")

# %%
# Historical Discount Factors
rates = []
rate = {"value_date": "2020-10-30", "st_busday": "Modified Following",
        "st_ratebasis": "Money Market", "st_daycount": "Actual/365",
        "lt_busday": "No Adjustment", "lt_frequency": "Semi-Annual",
        "lt_daycount": "Actual/Actual",
        "rates": {'O/N': 2.30, '1W': 2.35, '1M': 2.45, '3M': 2.55,
                  '6M': 2.65, '1Y': 2.70, '2Y': 2.80, '3Y': 2.90,
                  '5Y': 3.00, '10Y': 3.10, '30Y': 3.25}}
rates.append(rate)

rate = {"value_date": "2020-10-29", "st_busday": "Modified Following",
        "st_ratebasis": "Money Market", "st_daycount": "Actual/365",
        "lt_busday": "No Adjustment", "lt_frequency": "Semi-Annual",
        "lt_daycount": "Actual/Actual",
        "rates": {'O/N': 2.31, '1W': 2.34, '1M': 2.44, '3M': 2.56,
                  '6M': 2.67, '1Y': 2.75, '2Y': 2.85, '3Y': 2.95,
                  '5Y': 3.04, '10Y': 3.11, '30Y': 3.20}}
rates.append(rate)

rate = {"value_date": "2020-10-28", "st_busday": "Modified Following",
        "st_ratebasis": "Money Market", "st_daycount": "Actual/365",
        "lt_busday": "No Adjustment", "lt_frequency": "Semi-Annual",
        "lt_daycount": "Actual/Actual",
        "rates": {'O/N': 2.40, '1W': 2.44, '1M': 2.54, '3M': 2.66,
                  '6M': 2.77, '1Y': 2.85, '2Y': 2.95, '3Y': 3.05,
                  '5Y': 3.14, '10Y': 3.21, '30Y': 3.30}}
rates.append(rate)

rate = {"value_date": "2020-10-27", "st_busday": "Modified Following",
        "st_ratebasis": "Money Market", "st_daycount": "Actual/365",
        "lt_busday": "No Adjustment", "lt_frequency": "Semi-Annual",
        "lt_daycount": "Actual/Actual",
        "rates": {'O/N': 2.38, '1W': 2.44, '1M': 2.55, '3M': 2.76,
                  '6M': 2.87, '1Y': 3.05, '2Y': 3.15, '3Y': 3.25,
                  '5Y': 3.34, '10Y': 3.41, '30Y': 3.50}}
rates.append(rate)

rate = {"value_date": "2020-10-27", "st_busday": "Modified Following",
        "st_ratebasis": "Money Market", "st_daycount": "Actual/365",
        "lt_busday": "No Adjustment", "lt_frequency": "Semi-Annual",
        "lt_daycount": "Actual/Actual",
        "rates": {'O/N': 2.35, '1W': 2.40, '1M': 2.50, '3M': 2.65,
                  '6M': 2.75, '1Y': 2.85, '2Y': 2.95, '3Y': 3.05,
                  '5Y': 3.15, '10Y': 3.25, '30Y': 3.35}}
rates.append(rate)

dfs = dcurve.discount_factor_gen(rates, return_type="time")
# print(dfs)

# %%
# Define VaR Vertices using tenors
# convert the tenors to date then to no of days
from faspy.interestrate.rmp_dates import tenor_to_maturity as ttm, \
    day_count_factor as day_cf
var_vertices = ["6M", "1Y", "3Y", "5Y", "10Y", "20Y"]
var_dates = list(map(lambda tenor: ttm(vdate, tenor,
                                       business_day="Modified Following"),
                     var_vertices))
var_days = list(map(lambda date: (date - vdate).astype("int"), var_dates))
var_time = list(map(lambda date: day_cf("Actual/365", vdate, date),
                    var_dates))

# %%
# get historical discount factors
hdf = []
for df in dfs:
    x_axis = [x["times"] for x in df]
    y_axis = [y["df"] for y in df]
    idf = dcurve.interpolation(x_axis, y_axis, var_time)
    hdf.append(idf)
#print(hdf)

# %%
# 1.Calculate the log normal return
# 2. Calculate volatility
# 3. Calculate Correlation
from faspy.risk import var_utils as vu
from numpy import array as arr

returns = vu.returns_from_prices(arr(hdf))
vols = vu.volatilities(returns)
corr = vu.correlation(returns)
# print(returns, vols, corr)

# %%
# Bond Exposure
from faspy.interestrate.cashflows import fixbond_structures, fixbond_value
import numpy as np
from faspy.interestrate import rmp_dates as rd


mybond = {}
# mybond['issue_date'] = np.datetime64('2018-10-22')
mybond['value_date'] = np.datetime64('2020-10-22')
mybond['maturity'] = np.datetime64('2028-10-22')
mybond['day_count'] = 'Actual/365 Fixed'
mybond['frequency'] = 'Semi-Annual'
mybond['business_day'] = 'No Adjustment'
mybond['date_generation'] = rd.date_gen_method[1]
mybond['face_value'] = 1000000
mybond['coupon'] = 10
# mybond['ytm'] = 12.00
mybond['type'] = 'Fixed Rate Bond'

structures = list(fixbond_structures(mybond))
#print(structures)

# %%
# Reformat bond cash flows array

latest_df = dfs[0]
x_axis = [x["times"] for x in latest_df]
y_axis = [y["df"] for y in latest_df]
idf = dcurve.interpolation(x_axis, y_axis, var_days, is_function=True)
cf = []
for structure in structures:
    datum = {}
    if structure["end_date"] > vdate:
        datum["times"] = day_cf("Actual/365", vdate, structure["end_date"])
        datum["df"] = idf(datum["times"]) 
        datum["pv"] = datum["df"] * structure["cash_flow"]
        cf.append(datum)

# %%
# mapped bond cash flow to var vertices
vdf = hdf[0]
mcf = vu.map_cf_to_var_vertices(cf, vdf, var_time, corr, vols)
confidence = 0.99
# weigh the asset
weighted = vu.var_asset_weightbyprice(mcf, confidence, vols)
# calculate value at risk
va_risk = vu.value_at_risk(weighted, corr)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov  2 21:02:58 2020

@author: RMS671214
"""

from interestrate.fixincome import fixbond_structures, fixbond_value, \
    date_structures, fixbond
import numpy as np
from numpy import datetime64 as dt64
from interestrate import rmp_dates as rd
# %%

mybond = {}
mybond['issue_date'] = np.datetime64('2018-10-22')
mybond['value_date'] = np.datetime64('2021-10-22')
mybond['maturity'] = np.datetime64('2128-10-22')
mybond['day_count'] = 'Actual/365 Fixed'
mybond['frequency'] = 'Semi-Annual'
mybond['business_day'] = 'No Adjustment'
mybond['date_generation'] = rd.date_gen_method[1]
mybond['face_value'] = 1000000
mybond['coupon'] = 10.00
mybond['ytm'] = 10.00
mybond['type'] = 'Fixed Rate Bond'

structures = list(fixbond_structures(mybond))
print("STRUCTURES")
print("===========")
print(structures)

try:
    import pandas as pd
    pd1 = pd.DataFrame(structures)
except:
    pass


# %%

testdata = {"value_date":dt64("2020-05-01"), "maturity": dt64("2025-10-01"),
            "day_count": "Actual/Actual", "frequency":"Semi-Annual",
            "business_day": "No Adjustment",
            "date_generation": "Backward from maturity date",
            "face_value": 10000000, "coupon": 2.00, "ytm": 2.00}
val = fixbond(testdata)
print(val["risks"])


try:
    pd3 = pd.DataFrame(val["structure"])

except:
    import pandas as pd
    pd3 = pd.DataFrame(val["structure"])
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 12:05:24 2020

@author: RMS671214
"""

rate = {"value_date": str(mybond['value_date']),
        "st_busday": "Modified Following",
        "st_ratebasis": "Money Market", "st_daycount": "Actual/365",
        "lt_busday": "No Adjustment", "lt_frequency": "Semi-Annual",
        "lt_daycount": "Actual/Actual",
        "rates": {'O/N': 2.30, '1W': 2.35, '1M': 2.45, '3M': 2.55,
                  '6M': 2.65, '12M': 2.75, '1Y': 2.70, '2Y': 2.80, '3Y': 2.90,
                  '5Y': 3.00, '10Y': 3.10, '30Y': 3.25}}


dfs = dcurve.discount_factor_gen(rate, return_type="times")
x_axis = [x["times"] for x in dfs]
y_axis = [x["df"] for x in dfs]
df_func = rcurve.interpolation(x_axis, y_axis, float(1/366), is_function=True)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov  9 10:34:05 2020

@author: RMS671214
"""
from faspy.basel.credit.credit import calc_prob_default, \
    prob_default_interpolation as pdi, expected_loss, \
    mean_rr_with_betadist
        
import pandas as pd

# %%

spreads = []
spreads.append({"tenor": 1, "spread": 50, "rec_rate": 0.4})
spreads.append({"tenor": 2, "spread": 70, "rec_rate": 0.4})
spreads.append({"tenor": 3, "spread": 100, "rec_rate": 0.4})
spreads.append({"tenor": 5, "spread": 150, "rec_rate": 0.4})
spreads.append({"tenor": 10, "spread": 200, "rec_rate": 0.4})
spreads.append({"tenor": 20, "spread": 350, "rec_rate": 0.4})

print(spreads)
defaults = calc_prob_default(spreads)
pd_def = pd.DataFrame(defaults)

tenors = list(range(20))
print(tenors)
interp = pdi(defaults, tenors)
pd_interp = pd.DataFrame(interp)

# %%
# Test expected Loss

prob_default = 0.5
exposure_at_default = 1_000_000
rec_rate= 0.40
loss_given_default = None

eloss = expected_loss(prob_default, exposure_at_default, recovery_rate=rec_rate,
                      loss_given_default=loss_given_default)
print(eloss)

# %%
rr = mean_rr_with_betadist(0.5, 1)


import numba
from numpy import datetime64 as dt64
from .rmp_dates import day_count_factor as day_cf
from faspy.interestrate.rmp_curves import interpolation
import math


def bucketing(data, buckets, value_date, dfcurve=None):
    """
    Generate bonds coupon structures.

            Parameters:
                data: a list containing dictionary with the following keys -
                "id", "date" and "amount"
                buckets: a list containg dictionaries with the following keys -
                "name", "from" and "to". The value for each key is number of days


            Returns:
                a dictionary with the following keys - date, dcf, time,
                days, df, rate
    """
    vdate = dt64(value_date, "D")
    mydata = list(data)
    calc_data = [{"days": _datediff(vdate, dt64(x["date"], "D")).astype("int"),
                  "tenor": day_cf("Actual/365", vdate, dt64(x["date"], "D"))}
                 for x in mydata]
    if dfcurve is not None:
        df_xaxis = [x["times"] for x in dfcurve]
        df_yaxis = [x["df"] for x in dfcurve]
        i_func = interpolation(df_xaxis, df_yaxis, 1, is_function=True)
        df_array = [{"df": float(i_func(x["tenor"]))} for x in calc_data]


    dtlen = len(mydata)
    newdata = []
    for i in range(dtlen):
        old = dict(mydata[i])
        new = dict(calc_data[i])
        old.update(new)

        if dfcurve is not None:
            df_dic = df_array[i]
            old.update(df_dic)
            old["pv"] = old["df"] * old["amount"]
        newdata.append(old)

    mdur = [{"mod_duration": _mod_duration(float(x["tenor"]),
                                           float(x["df"]))}
            for x in newdata]
    bucket = [{"bucket": assign_bucket(x["tenor"], buckets)} for x in newdata]

    for i in range(dtlen):
        old = newdata[i]
        new = mdur[i]
        buc = bucket[i]
        old.update(new)
        old.update(buc)

    sortedbucket = sorted(buckets, key=sorter)
    buclen = len(sortedbucket)
    bucket_list = {}
    # return sortedbucket
    for i in range(buclen):
        bucket_list[sortedbucket[i]["name"]] = 0

    for datum in newdata:
        bucket_list[datum["bucket"]] += datum["pv"]

    return (newdata, bucket_list)


def sorter(item):
    return item["from"]


def assign_bucket(data, bucket):

    for buck in bucket:
        if data > buck["from"] and data <= buck["to"]:
            return buck["name"]

    return None


@numba.njit('float64(float64, float64)')
def _mod_duration(time, df):
    crate = -math.log(df)/time
    df01 = math.exp(-time * (crate + 0.0001))
    return (df01-df) / 0.0001


@numba.njit('float64(float64, float64)')
def _cal_crate(time, df):
    crate = -math.log(df)/time
    return float(crate)


@numba.njit
def _datediff(date1, date2):
    return date2 - date1



# !/usr/bin/env python3.
# -*- coding: utf-8 -*-
"""
Created on Sat Apr 25 00:08:48 2020

@author: ROSLI MOHD SANI
"""
from faspy.interestrate.rmp_dates import generate_datesv102 as gen_dates, \
    frequencies as fre
from faspy.interestrate.rmp_dates import day_count_factor as day_cf

from numpy import datetime64 as dt64, busday_offset as bus_off
import numpy as np
from faspy.interestrate.conventions import start_basis

import functools
from faspy.interestrate.rmp_solvers import solver_bond_derivatives as sbd, bond_risk
from numba import njit, jit


bondtype = ('Discount Bond', 'Convertible Discount Bond',
            'Callable Discount Bond', 'Callable Convertible Discount Bond',
            'Bullet Bond', 'Callable Bullet Bond',
            'Fixed Rate Bond', 'Fixed Rate ABS', 'Fixed Rate MBS',
            'Bond with Warrants', 'Convertible Bond',
            'Exchangeable Bond', 'Callable Bond', 'Callable ABS',
            'Callable MBS', 'Callable Convertible Bond',
            'Callable Exchangeable Bond', 'Amortising Bond',
            'Fixed Rate Amortising ABS Bond', 'Callable Amortising Bond',
            'Stepping FRB', 'Convertible Stepping Bond',
            'Exchangeable Stepping Bond', 'Callable Stepping Bond',
            'Callable Convertible Stepping Bond', 'Stepping Amortising Bond',
            'Callable Stepping Amortising Bond',
            'Floating Rate Note', 'Floating Rate ABS',
            'Floating Rate MBS', 'Floating Convertible Bond',
            'Floating Exchangeable Bond', 'Floating Callable Bond',
            'Floating Callable ABS', 'Floating Callable MBS',
            'Floating Callable Convertible Bond',
            'Floating Callable Exchangeable', 'Bond',
            'Floating Amortising Note', 'Stepping FRN',
            'Fixed Rate Bond', 'Discount Bond', 'Floating Rate Note',
            'Bond with Secondary Notes', 'Stepping Bonds with Secondary Notes',
            'Callable Convertible Discount Bond',
            'Callable Bond with Secondary Notes',
            'Callable Discount Bond', 'Callable Bond', 'Callable Bullet Bond',
            'Convertible Discount Bond', 'Convertible Bond',
            'Callable Convertible Stepping Bond',
            'Callable Stepping Bonds with Secondary Notes',
            'Callable Stepping Amortising Bonds with Secondary Notes',
            'Callable Stepping Bond', 'Amortising Bonds with Secondary Notes',
            'Bullet Bond', 'Stepping FRB', 'Callable Stepping Amortising Bond',
            'Stepping Amortising Bonds with Secondary Notes',
            'Floating Callable Bond', 'Convertible Stepping Bond',
            'Exchangeable Stepping Bond', 'Exchangeable Bond',
            'Callable Convertible Bond', 'Callable Exchangeable Bond',
            'Amortising Bond',
            'Callable Amortising Bond', 'Floating Amortising Note')
bondtype = set(bondtype)


# In[1]:

def construct_fixbond(bond):
    bus_day = None
    if bond['business_day'] == 'NULL':
        bus_day = 'No Adjustment'
    else:
        bus_day = bond['business_day']

    if bond['day_count'] == 'NULL':
        day_count = 'Actual/Actual ICMA'
    else:
        day_count = bond['day_count']

    start_date = bond.get('issue_date')
    if start_date is not None:
        start_date = dt64(start_date, 'D')
    value_date = bond.get('value_date')

    if value_date is not None:
        value_date = dt64(value_date, 'D')

    if start_date is not None and value_date is not None:
        use_date = start_date
    elif start_date is not None:
        use_date = start_date
    elif value_date is not None:
        use_date = value_date
    else:
        raise Exception('Both value_date and issue_date do not have any value')

    dates = gen_dates(use_date, bond['maturity'], issueDate=start_date,
                      frequency=fre[bond['frequency']],
                      business_day=bus_day, method=bond['date_generation'])

    #print(dates)
    # *********************************************************************************
    #
    #               CALCULATING THE DATA FOR FULL STRUCTURES
    #
    # *********************************************************************************
    fs = list(dates)
    flen = len(fs)
    for i in range(flen):
        fs_sgl = fs[i]
        fs_sgl['fv_flow'] = 0.00
        fs_sgl['fv'] = bond['face_value']
        if i > 0:
            fs_prev = fs[i-1]
            fs_sgl['cpn_dcf'] = day_cf(day_count, fs_prev['date'],
                                       fs_sgl['date'],
                                       bondmat_date=fs[-1]['date'],
                                       next_coupon_date=fs_sgl['date'],
                                       Frequency=12 /
                                       fre[bond['frequency']])
            fs_sgl['coupon'] = bond['coupon']
            fs_sgl['time'] = day_cf('Actual/365', use_date,
                                   fs_sgl['date'],
                                   bondmat_date=fs[-1]['date'],
                                   next_coupon_date=fs_sgl['date'],
                                   Frequency=12 /
                                   fre[bond['frequency']])
            fs_sgl['cf'] = (fs_sgl['coupon'] * fs_sgl['cpn_dcf'] * 0.01
                            * fs_sgl['fv'])
        else:
            fs_sgl['cpn_dcf'] = 0.00
            fs_sgl['coupon'] = 0.00
            fs_sgl['fv'] = 0.00
            fs_sgl['cf'] = 0.00

    fs[-1]['fv_flow'] = bond['face_value']
    bond['full_structures'] = fs

    # *********************************************************************************
    #
    #               CALCULATING THE DATA FOR ACTIVE STRUCTURES
    #
    # *********************************************************************************
    if value_date is None:
        bond['active_structures'] = {}

    elif value_date is not None and start_date is None:
        bond['active_structures'] = bond['full_structures']

    elif value_date < start_date:
        bond['active_structures'] = bond['full_structures']

    else:
        booldate = list(map(lambda x: x['date'] > value_date, fs))
        index = booldate.index(True)  # exclude paid coupon period/start_date
        acs = list(fs[index - 1:])
        acs[0]['dcf'] = 0
        acs[1]['dcf'] = day_cf(day_count, bond['value_date'], acs[1]['date'],
                               bondmat_date=acs[-1]['date'],
                               next_coupon_date=acs[1]['date'],
                               Frequency=12 / fre[bond['frequency']])
        alen = len(acs)
        for i in range(alen):
            acs[i]['time'] = day_cf('Actual/365', bond['value_date'],
                                    acs[i]['date'],
                                    bondmat_date=acs[-1]['date'],
                                    next_coupon_date=acs[i]['date'])

        bond['active_structures'] = acs


# In[]:
def construct_discountbond(bond):

    value_date = bond.get('value_date')
    issue_date = bond.get('issue_date')
    use_date = None
    if value_date is not None and issue_date is not None:
        if issue_date >= value_date:
            use_date = issue_date
        else:
            use_date = value_date
    elif value_date is not None:
        use_date = value_date
    elif issue_date is not None:
        use_date = issue_date

    dcfs = []
    times = []
    cfs = []
    cpns = []
    dates2 = []
    if use_date is not None:
        dcf = day_cf('Actual/365', use_date, bond['maturity'],
                     bondmat_date=bond['maturity'])
        dcfs.append(dcf)
        times.append(dcf)

    cfs.append(bond['position'])
    dates2.append(bond['maturity'])

    # cash_flows = {}
    cash_flows = {'dates': dates2, 'dcfs': dcfs, 'cfs': cfs, 'times': times}
    bond['cash_flows'] = cash_flows
    bond['coupon_structure'] = cpns
    # return bond


# %%
def construct_frn(bond, holidays=[]):
    bdc = np.busdaycalendar(weekmask='1111100', holidays=holidays)

    bus_day = None
    if bond['business_day'] == 'NULL':
        bus_day = 'No Adjustment'
    else:
        bus_day = bond['business_day']

    if bond['day_count'] == 'NULL':
        day_count = 'Actual/365'
    else:
        day_count = bond['day_count']

    start_date = bond.get('issue_date')
    if start_date is not None:
        start_date = dt64(start_date, 'D')
    value_date = bond.get('value_date')

    if value_date is not None:
        value_date = dt64(value_date, 'D')

    if start_date is not None and value_date is not None:
        use_date = start_date
    elif start_date is not None:
        use_date = start_date
    elif value_date is not None:
        use_date = value_date
    else:
        raise Exception('Both value_date and issue_date do not have any value')

    dates = gen_dates(use_date, bond['maturity'], issueDate=start_date,
                      frequency=fre[bond['frequency']],
                      business_day=bus_day, method=bond['date_generation'])

    #print(dates)
    # *********************************************************************************
    #
    #               CALCULATING THE DATA FOR FULL STRUCTURES
    #
    # *********************************************************************************
    fs = list(dates)

    flen = len(fs)
    for i in range(flen):
        row = fs[i]
        row['cf'] = 0
        row['time'] = day_cf('Actual/365', use_date, row['date'],
                               bondmat_date=fs[-1]['date'],
                               next_coupon_date=row['date'])
        if i == 0:
            row['dcf'] = 0
            row['y_dcf'] = 0
            row['margin'] = 0
            row['fv'] = 0
            row['fv_flow'] = 0
            row['fixing_date'] = None

        else:
            row['dcf'] = day_cf(day_count, fs[i-1]['date'], row['date'],
                                  bondmat_date=fs[-1]['date'],
                                  next_coupon_date=row['date'],
                                  Frequency=12 / fre[bond['frequency']])
            row['y_dcf'] = row['dcf']
            row['margin'] = bond['margin']
            row['fv'] = bond['face_value']
            row['fv_flow'] = 0
            if i == flen - 1:
                row['fv_flow'] = bond['face_value']
            row['fixing_date'] = bus_off(fs[i-1]['date'],
                                           -start_basis[bond['fixing_basis']],
                                           roll='backward', busdaycal=bdc)

    bond['full_structures'] = fs

    # *********************************************************************************
    #
    #               CALCULATING THE DATA FOR ACTIVE STRUCTURES
    #
    # *********************************************************************************
    if value_date is None:
        bond['active_structures'] = {}

    elif value_date is not None and start_date is None:
        bond['active_structures'] = bond['full_structures']

    elif value_date < start_date:
        bond['active_structures'] = bond['full_structures']

    else:
        # print(value_date,start_date)
        booldate = list(map(lambda x: x['date'] > value_date, fs))
        index = booldate.index(True)  # exclude paid coupon period/start_date
        acs = list(fs[index - 1:])
        acs[0]['y_dcf'] = 0
        acs[1]['y_dcf'] = day_cf(day_count, bond['value_date'], acs[1]['date'],
                               bondmat_date=acs[-1]['date'],
                               next_coupon_date=acs[1]['date'],
                               Frequency=12 / fre[bond['frequency']])
        alen = len(acs)
        for i in range(alen):
            acs[i]['time'] = day_cf('Actual/365', bond['value_date'],
                                    acs[i]['date'],
                                    bondmat_date=acs[-1]['date'],
                                    next_coupon_date=acs[i]['date'])

        bond['active_structures'] = acs

    return bond


# In[]:
def construct_bond(bond):

    if bond['type'] not in bondtype:
        raise Exception(bond['type'] + ' is not in the bondtype list')

    if bond['type'] == 'Fixed Rate Bond':
        construct_fixbond(bond)

    elif bond['type'] == 'FRN':
        pass

    elif bond['type'] == 'Discount Bond':
        construct_discountbond(bond)

    elif bond['type'] == 'Stepping Bond':
        pass


def fixbond_price(value_date, bond, yld):

    try:
        val_date = dt64(value_date)

    except Exception:
        raise Exception('Value date is not a date')

    try:
        acs = bond['active_structures']
    except Exception:
        raise Exception('full_structures does not have the required key(s)')

    try:
        freq = bond['frequency']
    except Exception:
        raise Exception('Error getting frequency of the floating leg')

    if bond['value_date'] is None:
        raise Exception('Value date for the bond is not set')

    bond['last_coupon_date'] = acs[0]['date']
    bond['accrued_interest'] = 0
    next_cpn = acs[1]
    # Calculation for fixed coupon
    if bond['value_date'] >= acs[0]['date']:
        coupon = bond.get('coupon')
        if coupon is not None:
            #print(next_cpn)
            next_cpn['coupon'] = coupon
            next_cpn['cf'] = coupon * next_cpn['dcf'] * next_cpn['fv'] * 0.01
            y_dcf = day_cf(bond['day_count'], val_date, next_cpn['date'],
                         bondmat_date=acs[-1]['date'],
                         next_coupon_date=next_cpn['date'],
                         Frequency=12 / fre[bond['frequency']])
            next_cpn['y_dcf'] = y_dcf
            df = 1 / (1 + yld * y_dcf * 0.01)
            next_cpn['pv'] = (next_cpn['cf'] + next_cpn['fv_flow']) * df
            next_cpn['df'] = df
            next_cpn['p_df'] = df

            # Calculation for PVBP01
            df01 = 1 / (1 + (yld + 0.01) * y_dcf * 0.01)
            next_cpn['pv01'] = (next_cpn['cf'] + next_cpn['fv_flow']) * df01
            next_cpn['df01'] = df01
            next_cpn['p_df01'] = df01
            #print(next_cpn)
        accrued_dcf = day_cf(bond['day_count'], bond['last_coupon_date'],
                             bond['value_date'], bondmat_date=acs[-1]['date'],
                             next_coupon_date=next_cpn['date'],
                             Frequency=12 / fre[bond['frequency']])
        # print(bond.get('current_coupon'), next_cpn)
        bond['accrued'] = accrued_dcf
        bond['accrued_interest'] = (accrued_dcf * next_cpn['coupon'] * 0.01 *
                                    next_cpn['fv'])

        lacs = len(acs)
        for i in range(2, lacs, 1):
            acs_sgl = acs[i]
            lastrow = acs[i-1]
            acs_sgl['y_dcf'] = acs_sgl['cpn_dcf']
            acs_sgl['p_df'] = 1 / (1 + yld * acs_sgl['y_dcf'] * 0.01)
            acs_sgl['p_df01'] = 1 / (1 + (yld + 0.01) * acs_sgl['y_dcf']
                                     * 0.01)
            acs_sgl['df'] = acs_sgl['p_df'] * lastrow['df']
            acs_sgl['df01'] = acs_sgl['p_df01'] * lastrow['df01']
            acs_sgl['pv'] = ((acs_sgl['cf'] + acs_sgl['fv_flow'])
                             * acs_sgl['df'])
            acs_sgl['pv01'] = ((acs_sgl['cf'] + acs_sgl['fv_flow'])
                               * acs_sgl['df01'])

        sum_pv = 0
        sum_pv01 = 0
        for i in range(1, lacs, 1):
            sgl_acs = acs[i]
            sum_pv += sgl_acs['pv']
            sum_pv01 += sgl_acs['pv01']
        bond['pvbp01'] = sum_pv01 - sum_pv
        bond['proceed'] = sum_pv
        bond['price per 100'] = ((bond['proceed'] - bond['accrued_interest'])
                                 / bond['face_value'] * 100)

        macD = 0
        for i in range(1, lacs, 1):
            acs_sgl = acs[i]
            macD = (macD + acs_sgl['time'] *
                    (acs_sgl['cf'] + acs_sgl['fv_flow']) *
                    acs_sgl['df'])

        macD = macD / bond['proceed']
        bond['modified duration'] = -macD / (1 + (yld * 0.01) / (12 / fre[freq]))
        bond['macaulay duration'] = macD
        #print('acs===>', acs)
        derivatives = bond_risk(acs, yld)
        bond['duration'] = derivatives['duration']
        bond['convexity'] = derivatives['convexity']
        bond['active_structures'] = acs
    else:
        # *******************************************************************
        # else clause process the bond as a forward issuance bond
        # disfac (discount factor) has to be provided to properly price the
        # bond) yld is taken to be the fwd yield of the bond
        # ********************************************************************
        raise Exception('Forward bond issuance is yet to be done')

# %%
def frn_price(bond, disc_curve):

    try:
        val_date = dt64(bond.get('value_date'))

    except Exception:
        raise Exception('Value date is not a date')

    try:
        acs = bond['active_structures']

    except Exception:
        raise Exception('full_structures does not have the required key(s)')

    bond['last_coupon_date'] = acs[0]['date']
    bond['accrued_interest'] = 0
    next_cpn = acs[1]
    interp = disc_curve.interpolate(1, is_function=True)
    # Calculation for fixed coupon
    if bond['value_date'] >= acs[0]['date']:
        coupon = next_cpn['coupon']
        if coupon is not None:
            gen_time = next_cpn['time']
            next_cpn['cf'] = coupon * next_cpn['dcf'] * next_cpn['fv'] * 0.01
            dcf = next_cpn['dcf']
            df = interp(gen_time)
            next_cpn['ref_rate'] = (1 / df - 1) / (dcf * 0.01)
            next_cpn['ref_with_spread'] = (next_cpn['ref_rate']
                                           + bond['spread'])
            df = 1 / (1 + next_cpn['ref_with_spread'] * dcf * 0.01)
            next_cpn['pv'] = (next_cpn['cf'] + next_cpn['fv_flow']) * df
            next_cpn['df'] = df
            # Calculation for PVBP01
            next_cpn['coupon01'] = coupon
            next_cpn['cf01'] = next_cpn['cf']
            df = interp(gen_time)
            next_cpn['ref_rate01'] = (1 / df - 1) / (dcf * 0.01) + 0.01
            next_cpn['ref01_with_spread'] = (next_cpn['ref_rate01']
                                             + bond['spread'])
            #print(next_cpn['ref_rate01'], bond['spread'])
            df01 = 1 / (1 + next_cpn['ref01_with_spread'] * dcf * 0.01)
            next_cpn['pv01'] = (next_cpn['cf01'] + next_cpn['fv_flow']) * df01
            next_cpn['df01'] = df01

    accrued_dcf = day_cf(bond['day_count'], bond['last_coupon_date'],
                         bond['value_date'], bondmat_date=acs[-1]['date'],
                         next_coupon_date=next_cpn['date'],
                         Frequency=12 / fre[bond['frequency']])
    #print(bond.get('current_coupon'), next_cpn)
    bond['accrued'] = accrued_dcf
    bond['accrued_interest'] = (accrued_dcf * next_cpn['coupon'] * 0.01 *
                                next_cpn['fv'])


    lacs = len(acs)
    # Calculation for unfixed coupon
    for i in range(2, lacs, 1):
        prev_acs = acs[i-1]
        sgl_acs = acs[i]
        #sgl_acs['fixed'] = False
        time0 = prev_acs['time']
        time1 = sgl_acs['time']
        df0 = disc_curve.interpolate(time0)
        df1 = disc_curve.interpolate(time1)
        dcf = sgl_acs['dcf']
        fwd_df = df1/df0
        sgl_acs['ref_rate'] = (1 / fwd_df - 1) / (dcf * 0.01)
        sgl_acs['coupon'] = (sgl_acs['ref_rate'] + sgl_acs['margin'])
        sgl_acs['cf'] = sgl_acs['coupon'] * dcf * 0.01 * sgl_acs['fv']
        sgl_acs['ref_with_spread'] = (sgl_acs['ref_rate'] + bond['spread'])
        sgl_acs['fwd_df'] = 1 / (1 + sgl_acs['ref_with_spread'] * dcf * 0.01)
        sgl_acs['df'] = sgl_acs['fwd_df'] * prev_acs['df']
        sgl_acs['pv'] = (sgl_acs['cf'] + sgl_acs['fv_flow']) * sgl_acs['df']
        # Calculation for PVBP01
        sgl_acs['ref_rate01'] = (1 / fwd_df - 1) / (dcf * 0.01) + 0.01
        sgl_acs['coupon01'] = (sgl_acs['ref_rate01'] + sgl_acs['margin'])
        sgl_acs['cf01'] = sgl_acs['coupon01'] * dcf * 0.01 * sgl_acs['fv']
        sgl_acs['ref01_with_spread'] = (sgl_acs['ref_rate01'] +
                                        bond['spread'])
        sgl_acs['fwd_df01'] = 1 / (1 + sgl_acs['ref01_with_spread'] *
                                   dcf * 0.01)
        sgl_acs['df01'] = sgl_acs['fwd_df01'] * prev_acs['df01']
        sgl_acs['pv01'] = ((sgl_acs['cf01'] + sgl_acs['fv_flow']) *
                           sgl_acs['df01'])

    sum_pv = 0
    sum_pv01 = 0
    for i in range(1, lacs, 1):
        sgl_acs = acs[i]
        sum_pv += sgl_acs['pv']
        sum_pv01 += sgl_acs['pv01']
    bond['pvbp01'] = sum_pv01 - sum_pv
    bond['proceed'] = sum_pv
    bond['price per 100'] = ((bond['proceed'] - bond['accrued_interest'])
                             / bond['face_value'] * 100)
    # derivatives = sbd(bond['face_value'], active_cfs, active_dcfs,
    # #                 active_fvs_flow, yld)
    # bond['duration'] = derivatives['duration']
    # bond['convexity'] = derivatives['convexity']



# %%
def bond_price(value_date, bond, freq=1):

    try:
        myyield = bond.get('ytm')
    except Exception:
        raise ValueError('ytm is required for valueation')

    bistype = bond.get('instrument_subtype')
    btype = bond.get('type')
    type_li = [bistype, btype]

    if bistype not in bondtype and btype not in bondtype:
        raise Exception(bistype + ' or ' + btype +
                        ' are not in the bondtype list')

    if 'Fixed Rate Bond' in type_li:
        fixbond_price(value_date, bond, myyield)

    elif 'Discount Bond' in type_li:
        cf = bond['cash_flows']
        bond['proceed'] = cf['cfs'][0] * (1 /
                                          (1 + (myyield * 0.01) /
                                           freq))**(cf['times'][0] * freq)
        value2 = cf['cfs'][0] * (1 /
                                 (1 + ((myyield + 0.01) * 0.01) /
                                  freq)) ** (cf['times'][0] * freq)
        bond['pvbp01'] = value2 - bond['proceed']
        bond['accrued interest'] = 0
        bond['price per 100'] = bond['proceed']/cf['cfs'][0] * 100
        bond['risk_stats'] = {}
        bond['risk_stats']['macaulay duration'] = cf['times'][0]
        bond['risk_stats']['modified duration'] = (cf['times'][0] /
                                                   (1 + (myyield * 0.01) /
                                                    (freq)))
        derivatives = sbd(bond['position'], cf['cfs'], cf['times'], myyield)
        bond['risk_stats']['duration'] = derivatives['duration']
        bond['risk_stats']['convexity'] = derivatives['convexity']

    else:
        raise Exception(btype + ' and ' + bistype + ' are not supported yet')


@jit
def _calc_df(active_cfs, active_dcfs, accrued_dcf, yld):
    dfs = []
    dfs01 = []
    totalcf = len(active_cfs)
    for i in range(totalcf):
        df = 1
        df01 = 1
        for ind in range(i+1):
            if ind == 0:
                df = df * 1 / (1 + yld *
                               (active_dcfs[ind] - accrued_dcf) * 0.01)
                df01 = df01 * 1 / (1 + (yld + 0.01) *
                                   (active_dcfs[ind] - accrued_dcf) * 0.01)
            else:
                df = df * 1 / (1 + yld * active_dcfs[ind] * 0.01)
                df01 = df01 * 1 / (1 + (yld + 0.01) * active_dcfs[ind] * 0.01)
        # print(df)
        dfs.append(df)
        dfs01.append(df01)

    return (dfs, dfs01)


@njit
def _calc_risk(active_cfs, dfs, dfs01, active_fvs_flow):
    # present value of each coupon flow
    dfs_len = len(dfs)
    active_cpns_values = []
    active_cpns_values01 = []
    active_pflow_values = []
    active_pflow_values01 = []
    for i in range(dfs_len):
        df = dfs[i]
        df01 = dfs01[i]
        cf = active_cfs[i]
        flow = active_fvs_flow[i]
        active_cpns_values.append(float(cf * df))
        active_pflow_values.append(float(flow * df))
        active_cpns_values01.append(float(cf * df01))
        active_pflow_values01.append(float(flow * df01))

    # net present value of all the principal flows
    pflow_value = 0.00
    pflow_len = len(active_pflow_values)
    for i in range(pflow_len):
        pflow_value = pflow_value + float(active_pflow_values[i])

    # summing the coupon pv to get the value of the bond
    cpn_values = 0.00
    cpn_len = len(active_cpns_values)
    for i in range(cpn_len):
        cpn_values = cpn_values + float(active_cpns_values[i])

    bond_values = cpn_values + pflow_value

    # net present value of all the principal flows
    pflow_value01 = 0.00
    pflow_len = len(active_pflow_values01)
    for i in range(pflow_len):
        pflow_value01 = pflow_value01 + float(active_pflow_values01[i])

    # summing the coupon pv to get the value of the bond
    cpn_values01 = 0.00
    cpn_len = len(active_cpns_values01)
    for i in range(cpn_len):
        cpn_values01 = cpn_values01 + float(active_cpns_values01[i])

    bond_values01 = cpn_values01 + pflow_value01
    pvbp01 = float(bond_values01 - bond_values)

    return (bond_values, pvbp01, active_cpns_values, active_pflow_values)



import numpy as np
from numpy import datetime64 as dt64
from .rmp_dates import tenor_to_maturity as ttm, \
day_count_factor as day_cf
from .rmp_dates import generate_dates as gen_dates
from scipy import interpolate
import sympy as sy
import numba
import math
import time
from collections import deque
from .conventions import frequencies



def generate_st_df(value_date, curve, day_count, business_day,
                   rate_basis='Money Market', holidays=[]):
    new_curve = []
    for k in curve:
        tenor = {}
        tenor['tenor'] = k
        tenor['rate'] = curve[k]
        tenor['date'] = ttm(value_date, k, convention=day_count,
                                business_day=business_day,
                                holidays=holidays)
        tenor['dcf'] = day_cf(day_count, value_date, tenor['date'])
        tenor['time'] = day_cf('Actual/365', value_date, tenor['date'])
        tenor["days"] = (tenor["date"] - value_date).astype("int")
        if rate_basis == 'Money Market':
            tenor['df'] = _mmr2df(tenor['rate'], tenor['dcf'])
        else:
            tenor['df'] = _dr2df(tenor['rate'], tenor['dcf'])
        new_curve.append(tenor)

    return new_curve


def generate_st_df_bymaturity(value_date, maturity, rate, convention,
                              business_day, rate_basis='Money Market',
                              holidays=[]):
    result = {}
    dcf = day_cf(convention, value_date, maturity)

    time = day_cf('Actual/365', value_date, maturity)
    #print('dcf: ',dcf)
    df = None
    if rate_basis == 'Money Market':
        df = 1/(1 + rate * 0.01 * dcf)
    elif rate_basis == 'Discount Rate':
        df = 1 - rate * 0.01 * dcf
    #print('dfs in generate_st_df: ', dfs)

    result['dates'] = maturity
    result['dcfs'] = dcf
    result['dfs'] = df
    result['times'] = time

    return result


# In[10]:
def generate_fulldf(value_date, st_curve, st_daycount, st_business_day,
                    st_rate_basis, lt_curve, lt_daycount,
                    lt_business_day, frequency=6,
                    method="Forward from issue date",
                    holidays=[]):
    par_time = deque()
    par_rate = deque()
    df_time = deque()
    df_df = deque()
    st_ = generate_st_df(value_date, st_curve, st_daycount,
                         st_business_day, rate_basis=st_rate_basis)

    len_st = len(st_)
    st2compound = convert_shortrate_to_compounding
    for i in range(len_st):
        lst_ = st_[i]
        lst_['par_rate'] = st2compound(lst_['rate'], value_date,
                                       lst_['date'],
                                       frequency=12,
                                       compound_busday=lt_business_day,
                                       compound_dc=lt_daycount,
                                       compound_frequency=frequency)
        par_time.append(lst_['time'])
        par_rate.append(lst_['par_rate'])
        df_time.append(lst_['time'])
        df_df.append(lst_['df'])

    lt_ = deque()

    for k in lt_curve:
        tenor = {}
        tenor['tenor'] = k
        tenor['rate'] = lt_curve[k]
        tenor['date'] = ttm(value_date, k, convention=lt_daycount,
                                business_day=lt_business_day,
                                holidays=holidays)
        tenor['time'] = day_cf('Actual/365', value_date, tenor['date'])
        tenor["days"] = (tenor["date"] - value_date).astype("int")
        time_bool = list(map(lambda ptime: tenor['time'] >= ptime, par_time))

        try:
            index = time_bool.index(False)
            if tenor['time'] not in par_time:
                par_time.insert(index, tenor['time'])
                par_rate.insert(index, tenor['rate'])
        except ValueError:
            # print(tenor['time'] not in par_time, tenor['time'])
            if tenor['time'] not in par_time:
                par_time.append(tenor['time'])
                par_rate.append(tenor['rate'])
        lt_.append(tenor)
        # print(tenor)

    # print(lt_)
    lt_last_maturity = lt_[-1]['date']

    lt_alldates = gen_dates(value_date, lt_last_maturity, issueDate=value_date,
                            frequency=frequency, business_day=lt_business_day,
                            method=method, holidays=[])
    len_dates = len(lt_alldates)
    lt_all = deque()
    interp_par = interpolation(par_time, par_rate, 1, model='chip',
                               is_function=True)
    for i in range(len_dates):
        lt_sgl = {}
        thedate = lt_alldates[i]
        if thedate != value_date:
            lt_sgl['date'] = thedate
            lt_sgl["days"] = (thedate - value_date).astype("int")
            lt_sgl['time'] = day_cf('Actual/365', value_date, thedate)
            if i != len_dates - 1:
                lt_sgl['dcf'] = day_cf(lt_daycount, lt_alldates[i-1],
                                       thedate, bondmat_date=lt_alldates[-1],
                                       next_coupon_date=lt_alldates[i+1])
            else:
                lt_sgl['dcf'] = day_cf(lt_daycount, lt_alldates[i-1],
                                       thedate, bondmat_date=lt_alldates[-1],
                                       next_coupon_date=lt_alldates[-1])
            lt_sgl['rate'] = float(interp_par(lt_sgl['time']))
            lt_all.append(lt_sgl)

    # *********************************************************************
    #   BOOTSTRAPPING STARTS HERE
    # *********************************************************************
    len_all = len(lt_all)
    lt_all = list(lt_all)
    for i in range(len_all):
        lt_sgl = lt_all[i]
        ctime = lt_sgl['time']
        if ctime <= max(df_time):
            idf = interpolation(df_time, df_df, ctime)
            lt_sgl['df'] = idf
        else:

            dcfs = np.array([x['dcf'] * x['df'] for x in lt_all[:i]])
            new_dcfs = dcfs * (lt_sgl['rate'] * 0.01)
            values = np.sum(new_dcfs)
            cdf = (1 - values)/(1 + lt_sgl['rate'] * 0.01 * lt_sgl['dcf'])
            lt_sgl['df'] = cdf
            df_time.append(lt_sgl['time'])
            df_df.append(cdf)

    # merging the discount factors
    new_df = [*st_, *lt_all]
    def sorter(items):
        return items['time']

    new_df = sorted(new_df, key=sorter)
    lnew_df = len(new_df)
    dfs = deque()
    dfs.append(new_df[0])
    for i in range(1, lnew_df, 1):
        sgl_newdf = new_df[i]
        if sgl_newdf['time'] != new_df[i-1]['time']:
            dfs.append(sgl_newdf)

    #new_df = list(set(new_df))
    # return lt_all
    return dfs


def interpolation(x_axis, y_axis, x_value, model='chip', method=None,
                  is_function=False):
    methods = ['linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic',
               'previous', 'next']
    # print(type(x_value),type(min(x_axis)))
    models = ['akima', 'chip', 'interp1d', 'cubicspline', 'krogh',
              'barycentric']
    # print('x_axis => ',x_axis)

    result = None
    f = None
    if model is None or model == 'interp1d':
        if method in methods:
            f = interpolate.interp1d(x_axis, y_axis, kind=method)
        else:
            f = interpolate.interp1d(x_axis, y_axis, kind='cubic')

    elif model in models:
        if model == 'akima':
            f = interpolate.Akima1DInterpolator(x_axis, y_axis)

        elif model == 'chip':
            f = interpolate.PchipInterpolator(x_axis, y_axis)

        elif model == 'cubicspline':
            f = interpolate.CubicSpline(x_axis, y_axis)

        elif model == 'krogh':
            f = interpolate.KroghInterpolator(x_axis, y_axis)

        elif model == 'barycentric':
            f = interpolate.BarycentricInterpolator(x_axis, y_axis)
    else:
        f = interpolate.PchipInterpolator(x_axis, y_axis)

    if is_function == True:
        return f
    else:
        if not isinstance(x_value, list):
            # if x_value <min(x_axis) or x_value >max(x_axis):
            #    raise Exception('interpolation error: value requested is outside of range')
            #    return result
            try:
                result = float(f(x_value))
            except:
                return result

        else:
            result = list(map(lambda x: float(f(x)),x_value))

        return result


def convert_shortrate_to_compounding(rate, start, end, frequency=12,
                                     convention='Actual/365',
                                     compound_busday='No Adjustment',
                                     rate_basis='Money Market',
                                     compound_dc='Actual/365',
                                     compound_frequency=12,
                                     method="Forward from issue date"):
    result = None
    fre = 12/frequency
    compound_fre = 12/compound_frequency

    dcf = day_cf(convention, start, end, Frequency=fre)
    df = None
    if rate_basis == 'Money Market':
        df = 1/(1 + rate * 0.01 * dcf)
    else:
        df = 1-rate * 0.01 * dcf
    two_year_date = ttm(start, '2Y',
                                          convention=compound_dc,
                                          business_day=compound_busday,
                                          holidays=[])
    li_cpn_dcf = []

    dates = gen_dates(start, two_year_date, issueDate=start,
                               frequency=compound_frequency,
                               business_day=compound_busday,
                               method=method, holidays=[])

    no_of_cpn = len(dates)

    for cpn_no in range(no_of_cpn-1):
        prev_cpn = dates[cpn_no]
        next_cpn = dates[cpn_no +1]
        if end <= prev_cpn:
            break
        elif next_cpn >= end:
            cur_date = end
        else:
            cur_date = next_cpn

        cmp_dcf = day_cf(compound_dc,prev_cpn, cur_date,
                               bondmat_date = dates[-1], next_coupon_date = next_cpn,
                               business_day = compound_busday,
                               Frequency = compound_fre)

        #print('start=>',start,'end=>',end,'cpn_no==>',cpn_no,'prev_cpn=>',prev_cpn,'cur_date==>', cur_date, 'nextcpn==>', next_cpn)
        #print('cmp_dcf==>>', cmp_dcf)
        li_cpn_dcf.append(cmp_dcf)

    #print('dates==>',dates,'enddate===>',end,'startdate===>',start,
    #      'df:', df,'li:',li_cpn_dcf)
    result = solver_rate_from_compounded_df(df,li_cpn_dcf)
    #print(type(result),len(result),result)
    result = float(result[0])

    #print(start,end,result)

    return result


def solver_rate_from_compounded_df(dis_factor,daycount_factors):
    rate = sy.Symbol('rate')
    fv = sy.Symbol('fv')
    #df = sy.Symbol('df')
    dcfs = daycount_factors
    #df = dis_factor

    for d in range(len(dcfs)):
        if d == 0:

            fv = (1 + (rate* 0.01*dcfs[d]) )
        else:
            fv = fv * (1 + (rate* 0.01*dcfs[d]) )

    fv = dis_factor * fv - 1
    #solved_rates = sy.solve(fv,rate)
    solved_rates = sy.solveset(fv, rate, domain=sy.S.Reals)
    new_rates = []
    #print('solved_rates==> ',solved_rates, type(solved_rates),list(solved_rates))
    solved_rates = list(solved_rates)
    for i in range(len(solved_rates)):
        try:
            float(solved_rates[i])
        except:
            continue
        new_rates.append(solved_rates[i])

    new_rates = list(filter(lambda x: x>0, new_rates))
    return new_rates


def calc_fwd_df(start, end, time_axis=None, df_axis=None, ifunc=None):

    if ifunc is not None:
        df1 = ifunc(start)
        df2 = ifunc(end)
        return df2 / df1

    elif (time_axis is not None and df_axis is not None):
        interp = interpolation(time_axis, df_axis, start, is_function=True)
        df1 = interp(start)
        df2 = interp(end)
        return df2 / df1
    else:
        return None


def calc_shortrate_from_df(startdate, enddate, df, day_count,
                           rate_basis="Money Market"):
    sdate = dt64(startdate, "D")
    edate = dt64(enddate, "D")
    dcf = day_cf(day_count, sdate, edate, bondmat_date=edate,
                 next_coupon_date=edate)
    if rate_basis == "Discount Rate":
        return _df2dr(df, dcf)
    else:
        return _df2mmr(float(df), float(dcf))
    return None


def calc_df_from_shortrate(startdate, enddate, rate, day_count,
                           rate_basis="Money Market"):
    sdate = dt64(startdate, "D")
    edate = dt64(enddate, "D")
    dcf = day_cf(day_count, sdate, edate, next_coupon_date=edate)
    if rate_basis == "Discount Rate":
        return _dr2df(rate, dcf)
    else:
        return _mmr2df(rate, dcf)
    return None


def continuous_rate(value_date, rate, tenor, day_count="Actual/365",
                    rate_basis="Money Market"):
    maturity = ttm(value_date, tenor, convention=day_count)
    time = day_cf("Actual/365", value_date, maturity)
    dcf = day_cf(day_count, value_date, maturity)
    if rate_basis == "Money_Market":
        df = _mmr2df(float(rate), float(dcf))
    elif rate_basis == "Discount Rate":
        df = _dr2df(float(rate), float(dcf))

    crate = -math.log(df)/time
    return crate


def shift_curve(curve, bp=0.01):
    
    rates = dict(curve)
    for key in rates:
        rates[key] += bp
    return rates


def discount_factor_from_zspread(value_date, date_structure, day_count,
                                 frequency, df_curve, zspread):
    data = [{"start_date": x["start_date"], "end_date": x["end_date"]}
            for x in date_structure if value_date < x["end_date"]]

    x_axis = [x["times"] for x in df_curve]
    y_axis = [x["df"] for x in df_curve]
    ifunc = interpolation(x_axis, y_axis, 1/366, is_function=True)
    maturity = data[-1]["end_date"]
    zdfs = [{"times": 0, "df": 1}]
    for datum in data:
        if value_date > datum["start_date"]:
            time1 = day_cf("Actual/365", value_date, datum["end_date"],
                           bondmat_date=maturity,
                           next_coupon_date=datum["end_date"])
            df = ifunc(time1)
            fwd_rate = calc_shortrate_from_df(value_date,
                                              datum["end_date"],
                                              df, day_count)
            zfwd_rate = fwd_rate + zspread
            zfwd_df = calc_df_from_shortrate(value_date, datum["end_date"],
                                             zfwd_rate, day_count)
            zdfs.append({"times": time1, "zfwd_df": zfwd_df})

        else:
            start_time = day_cf("Actual/365", value_date,
                                datum["start_date"],
                                bondmat_date=maturity,
                                next_coupon_date=datum["end_date"])
            end_time = day_cf("Actual/365", value_date, datum["end_date"],
                              bondmat_date=maturity,
                              next_coupon_date=datum["end_date"])
            fwd_df = calc_fwd_df(start_time, end_time, ifunc=ifunc)
            fwd_rate = calc_shortrate_from_df(datum["start_date"],
                                              datum["end_date"],
                                              fwd_df, day_count)

            zfwd_rate = fwd_rate + zspread
            zfwd_df = calc_df_from_shortrate(datum["start_date"],
                                             datum["end_date"],
                                             zfwd_rate, day_count)
            zdfs.append({"times": end_time, "zfwd_df": zfwd_df})

    data_len = len(zdfs)

    for i in range(1, data_len, 1):
        
        datum = zdfs[i]
        pdatum = zdfs[i-1]
        datum["df"] = pdatum["df"] * datum["zfwd_df"]

    zdf_curve = [{"times": x["times"], "df": x["df"]} for x in zdfs]
    

    return zdf_curve


@numba.njit('float64(float64, float64)')
def _df2mmr(df, dcf):
    rate = ((1 / df) - 1) / (0.01 * dcf)
    return rate


@numba.njit('float64(float64, float64)')
def _mmr2df(rate, dcf):
    df = 1 / (1 + rate * 0.01 * dcf)
    return df


@numba.njit('float64(float64, float64)')
def _df2dr(df, dcf):
    rate = (1 - df) / (0.01 * df)
    return rate


@numba.njit('float64(float64, float64)')
def _dr2df(rate, dcf):
    df = 1 - rate * 0.01 * dcf
    return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul  7 17:33:15 2020

@author: RMS671214
"""
std_tenors =('1W', '1M', '3M', '6M', '12M', '1Y', '2Y', '3Y', '5Y', '10Y', '30Y')
rate_basises = ("Simple", "Disount Rate", "Continuous")
convs = ('Actual/365', 'Actual/365 Fixed', 'Actual/365L', 'Actual/364',
         'Actual/360', 'Eurobond basis (ISDA 2006)', '30A/360', '30/360 US',
         '30/360 ICMA', 'Eurobond basis (ISDA 2000)', 'Actual/Actual',
         'Actual/Actual ISDA', 'Actual/Actual AFB')  # 'Actual/Actual ICMA'
day_counts = convs
all_cons = []
for conv in convs:
    tup_conv = (conv, conv)
    all_cons.append(tup_conv)


business_days = ("No Adjustment", "Following", "Modified Following",
                 "Preceeding", "Modified Preceeding", "End Of Month")

date_gen_method = ("Forward from issue date", "Backward from maturity date")
frequencies = {"Annual": 12, "Semi-Annual": 6, "Quarterly": 3, "Monthly": 1}
start_basis = {'Spot': 2, 'Tom': 1, 'Same Day': 0}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov  2 21:02:58 2020

@author: RMS671214
"""

from interestrate.fixincome import date_structures, calc_customfix_structures, \
    value_customfix_structures, create_structures_from_dates
import numpy as np
from interestrate import rmp_dates as rd
from interestrate import discount_curve as dcurve
# %%

mybond = {}
mybond['issue_date'] = np.datetime64('2018-10-22')
mybond['value_date'] = np.datetime64('2021-10-22')
mybond['maturity'] = np.datetime64('2028-10-22')
mybond['day_count'] = 'Actual/365 Fixed'
mybond['frequency'] = 'Semi-Annual'
mybond['business_day'] = 'No Adjustment'
mybond['date_generation'] = rd.date_gen_method[1]
mybond['face_value'] = 1000000
mybond['coupon'] = 10
mybond['ytm'] = None
mybond['type'] = 'Fixed Rate Bond'

dates = list(date_structures(mybond))
print("DATES")
print("=====")
print(dates)

for date in dates:
    date["start_date"] = str(date["start_date"])
    date["end_date"] = str(date["end_date"])
    
print("DATES AS STRING")
print("===============")
print(dates)

try:
    import pandas as pd
    pd1 = pd.DataFrame(dates)
except:
    pass


# %%
# Create Sample structure

lenstruc = len(dates)
face_values = [10_000_000.00  + i * 100_000.00 for i in range(lenstruc)]
print("FACE VALUES")
print("===========")
print(face_values)
coupons = [3.5  + i * 0.20 for i in range(lenstruc)]
print("COUPONS")
print("=======")
print(coupons)

fv_flows = [i * 100_000.00 for i in range(lenstruc)]
print("FV FLOWS")
print("========")
print(fv_flows)
fv_flows[-1] = -face_values[-1]

new_structures = create_structures_from_dates(dates, coupons, face_values, fv_flows)
print("CUSTOM STRUCTURE")
print("================")
print(new_structures)
try:
    import pandas as pd
    pd2 = pd.DataFrame(new_structures)
except:
    pass

# %% value structure using YTM

astructures = value_customfix_structures(mybond["value_date"], new_structures,
                                         mybond["day_count"],
                                         mybond["frequency"],
                                         5.00)

print("VALUATION OF CUSTOM STRUCTURE")
print("=============================")
print(astructures)
try:
    import pandas as pd
    pd3 = pd.DataFrame(astructures)
except:
    pass

# %% value struture using disount factors

rate = {"value_date": "2020-10-30", "st_busday": "Modified Following",
        "st_ratebasis": "Money Market", "st_daycount": "Actual/365",
        "lt_busday": "No Adjustment", "lt_frequency": "Semi-Annual",
        "lt_daycount": "Actual/Actual",
        "rates": {'O/N': 2.30, '1W': 2.35, '1M': 2.45, '3M': 2.55,
                  '6M': 2.65, '12M': 2.75, '1Y': 2.70, '2Y': 2.80, '3Y': 2.90,
                  '5Y': 3.00, '10Y': 3.10, '30Y': 3.25}}

dfs = dcurve.discount_factor_gen(rate, return_type="time")

astructures1 = value_customfix_structures(mybond["value_date"], new_structures,
                                         mybond["day_count"],
                                         mybond["frequency"],
                                         dfs)

print("VALUATION OF CUSTOM STRUCTURE")
print("=============================")
print(astructures)
try:
    import pandas as pd
    pd4 = pd.DataFrame(astructures1)
except:
    pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr 26 14:11:55 2020

@author: RMS671214
"""
from sympy import Symbol, diff, solve


def solver_bond_derivatives(fv, cfs=[], yld_dcfs=[], fvs=[], my_yield=5):
    yld = Symbol('yld')
    df = Symbol('df')
    expr = Symbol('expr')
    expr = None
    dcfs_len = len(yld_dcfs)
    for i in range(dcfs_len):
        ci = None
        df = None
        if yld_dcfs[i] != 0:
            ci = (cfs[i] + fvs[i]) / fv * 100

            if df is None:
                df = (1 + yld * yld_dcfs[i] * 0.01)

            for k in range(i):
                if yld_dcfs[k] != 0:
                    df = df * (1 + yld * yld_dcfs[k] * 0.01)

            if expr is None:
                expr = ci/df
            else:
                expr = expr + ci/df

    expr_duration = diff(expr, yld)
    duration = expr_duration.evalf(subs={yld: my_yield})
    # print(type(duration))
    # print('Duration >>>>',duration)
    expr_convex = diff(expr_duration, yld)
    convexity = expr_convex.evalf(subs={yld: my_yield})
    # print('Convexity >>>>>>',convexity)
    return {'duration': float(duration), 'convexity': float(convexity)}


def bond_risk(structure, my_yield=5):
    yld = Symbol('yld')
    df = Symbol('df')
    expr = Symbol('expr')
    expr = None
    dcfs_len = len(structure)
    fv = structure[-1]['fv']
    #print('structure ==>>', structure)
    for i in range(1, dcfs_len, 1):
        row = structure[i]
        #print(row)
        ci = None
        df = None
        #print('i======>>>>', i, row)
        if row['time'] > 0:
            ci = (row['cf'] + row['fv_flow']) / fv * 100

            if df is None:
                df = (1 + yld * row['y_dcf'] * 0.01)

            for k in range(1, i, 1):
                row2 = structure[k]
                if row2['y_dcf'] != 0:
                    df = df * (1 + yld * row2['y_dcf'] * 0.01)

            if expr is None:
                expr = ci/df
            else:
                expr = expr + ci/df

    expr_duration = diff(expr, yld)
    duration = expr_duration.evalf(subs={yld: my_yield})
    # print(type(duration))
    # print('Duration >>>>',duration)
    expr_convex = diff(expr_duration, yld)
    convexity = expr_convex.evalf(subs={yld: my_yield})
    # print('Convexity >>>>>>',convexity)
    return {'duration': float(duration), 'convexity': float(convexity)}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov  4 16:20:19 2020

@author: RMS671214
"""

from faspy.interestrate.rmp_stats import rmpcdf, rmpppf
import math
from numba import njit
from faspy.interestrate import rmp_curves as rcurve
from collections import deque

def expected_loss(prob_default, exposure_at_default, recovery_rate=None,
                  loss_given_default=None):
    """
    Calculate the expected loss.

    Parameters
    ----------
    prob_default : float
        Probability of default.
    exposure_at_default : float
        Exposure at default.
    recovery_rate : float
        Recovery rate. Defaulted at None. If None 'loss_given_default' must be provided
    loss_given_default : float
        Loss given default. Defaulted at None. If None, recovery rate must be provided.

    Returns
    -------
    float
        Expected Loss.

    """
    try:
        rr = float(recovery_rate)
        try:
            el = prob_default * (1 - rr) * exposure_at_default
            return el
        except Exception as exc:
            return str(exc)
    except:
        try:
            lgd = float(loss_given_default)
            try:
                el = prob_default * lgd * exposure_at_default
                return el
            except Exception as exc:
                return str(exc)
        except Exception as exc:
            return str(exc)


def mean_rr_with_betadist(alpha, beta):
    return alpha / (alpha + beta)


def basel_cap_req(loss_given_default, prob_default, corr, confidence=0.999,
                 mat_adj=1):
    d1 = (rmpppf(prob_default) * math.sqrt(1 / (1 - corr)) +
          rmpppf(confidence) * math.sqrt(corr / (1 - corr)))
    cond_exp_loss = loss_given_default * rmpcdf(d1)
    exp_loss = loss_given_default * prob_default
    cap_req = (cond_exp_loss - exp_loss) * mat_adj
    return cap_req


def maturity_adjustment(prob_default, maturity, fixed_maturity=2.5):
    b_pd = math.pow(0.11852 - 0.05478 * math.log(prob_default), 2)
    mat_adj = (1 + (maturity - fixed_maturity) + b_pd) / (1 - 1.5 * b_pd)
    return mat_adj


def rwa(exp_at_default, car, cap_req):
    return cap_req * exp_at_default / car


def credit_correlation(corr_lo_pd, corr_hi_pd, prob_default, kfactor,
                       size_adj=0):
    factor_weight = ((1 - math.exp(-kfactor * prob_default)) /
                     (1 - math.exp(-kfactor)))
    w1 = corr_hi_pd * factor_weight
    w2 = corr_lo_pd * factor_weight

    return w1 + w2 - size_adj


def size_adjustment(sales, min_sales=5, max_sales=50):
    if sales <= min_sales:
        return 0.04
    elif sales >= max_sales:
        return 0.00s
    else:
        return 0.04 * (1 - (sales - min_sales)/(max_sales - min_sales))


def calc_prob_default(data):
    """
    Calculate probability of default and its cumulative.

            Parameters:
                data: is an array of dictionary. The dictionary must have the
                following keys - tenor, spread, rec_rate.

            Returns:
                an array of dictionary with the following keys - tenor, spread
                prob_default and cum_prob_default
    """
    datas = list(data)
    for datum in data:
        datum["prob_default"] = 1 - math.exp((-datum["spread"]) /
                                             ((1 - datum["rec_rate"]) * 10000))
        datum["cum_prob_default"] = 1 - math.exp((-datum["spread"] *
                                                  datum["tenor"]) /
                                                 ((1 - datum["rec_rate"]) *
                                                  10000))

    return datas


def prob_default_interpolation(defaults, tenors):
    """
    Interpolation of default probability

    Parameters
    ----------
    defaults : [{"tenor":float, "prob_default": float, "cum_prob_default": float}]
        The probability default curve
    tenors : [float]
        Tenor stated in the same manner as "tenor" key of "defaults" e.g time as fraction of year or number of days.

    Returns
    -------
    [float]
        Default probability for each tenor.

    """
    x_axis = [x["tenor"] for x in defaults]
    x_axis.insert(0, 0)

    def_y_axis = [x["prob_default"] for x in defaults]
    def_y_axis.insert(0, 0)

    cum_def_y_axis = [x["cum_prob_default"] for x in defaults]
    cum_def_y_axis.insert(0, 0)

    def_func = rcurve.interpolation(x_axis, def_y_axis, 1, is_function=True)
    cum_def_func = rcurve.interpolation(x_axis, cum_def_y_axis, 1, is_function=True)

    results = deque()
    for tenor in tenors:
        result = {}
        result["tenor"] = tenor
        result["prob_default"] = def_func(tenor)
        result["cum_prob_default"] = cum_def_func(tenor)
        results.append(result)
    return list(results)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov  2 21:02:58 2020

@author: RMS671214
"""

from faspy.interestrate.cashflows import fixbond_structures, fixbond_value, \
    coupon_structures
import numpy as np
from faspy.interestrate import rmp_dates as rd
# %%

mybond = {}
mybond['issue_date'] = np.datetime64('2018-10-22')
mybond['value_date'] = np.datetime64('2021-10-22')
mybond['maturity'] = np.datetime64('2028-10-22')
mybond['day_count'] = 'Actual/365 Fixed'
mybond['frequency'] = 'Semi-Annual'
mybond['business_day'] = 'No Adjustment'
mybond['date_generation'] = rd.date_gen_method[1]
mybond['face_value'] = 1000000
mybond['coupon'] = 10
mybond['ytm'] = 12.00
mybond['type'] = 'Fixed Rate Bond'

structures = list(fixbond_structures(mybond))
print(structures)

# %%
fixbond_value(mybond["value_date"], structures, mybond["ytm"],
              mybond["day_count"], mybond["frequency"])
print(structures)
value = 0
accrued = 0
for structure in structures:
    value += structure["pv"]
    accrued += structure["accrued"]
    
print("value: ", value)
print("accrued: ", accrued)

# %%

coupon_dates = list(coupon_structures(mybond))
print(coupon_dates)